import operator as op
import typing as ta
import weakref

from .. import lang


T = ta.TypeVar('T')
K = ta.TypeVar('K')
V = ta.TypeVar('V')


##


class IdentityKeyDict(ta.MutableMapping[K, V]):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()

        self._dict: dict[int, tuple[K, V]] = {}
        for k, v in lang.yield_dict_init(*args, **kwargs):
            self[k] = v

    def __reduce__(self):
        return (type(self), (list(self.items()),))

    @property
    def debug(self) -> ta.Sequence[tuple[K, V]]:
        return list(self.items())

    def __repr__(self) -> str:
        return lang.attr_repr(self, '_dict')

    def __setitem__(self, k: K, v: V) -> None:
        self._dict[id(k)] = (k, v)

    def __delitem__(self, k: K) -> None:
        del self._dict[id(k)]

    def __getitem__(self, k: K) -> V:
        return self._dict[id(k)][1]

    def __len__(self) -> int:
        return len(self._dict)

    def __iter__(self) -> ta.Iterator[K]:
        return iter(map(op.itemgetter(0), self._dict.values()))

    def clear(self) -> None:
        self._dict.clear()


class IdentitySet(ta.MutableSet[T]):
    def __init__(self, init: ta.Iterable[T] | None = None) -> None:
        super().__init__()

        self._dict: dict[int, T] = {}
        if init is not None:
            for item in init:
                self.add(item)

    def __reduce__(self):
        return (type(self), (list(self),))

    @property
    def debug(self) -> ta.Sequence[T]:
        return list(self)

    def __repr__(self) -> str:
        return lang.attr_repr(self, '_dict')

    def add(self, item: T) -> None:
        self._dict[id(item)] = item

    def discard(self, item: T) -> None:
        try:
            del self._dict[id(item)]
        except KeyError:
            pass

    def update(self, items: ta.Iterable[T]) -> None:
        for item in items:
            self.add(item)

    def __contains__(self, item: T) -> bool:  # type: ignore
        return id(item) in self._dict

    def __len__(self) -> int:
        return len(self._dict)

    def __iter__(self) -> ta.Iterator[T]:
        return iter(self._dict.values())


##


class IdentityWeakKeyDictionary(ta.MutableMapping[K, V]):
    """
    https://ideone.com/G4iIri
    https://stackoverflow.com/questions/75314250/python-weakkeydictionary-for-unhashable-types#comment135919973_77100606

    See also:
    https://github.com/python-trio/trio/blob/efd785a20721707b52a6e2289a65e25722b30c96/src/trio/_core/_ki.py#L81

    TODO:
     - audit for freethreaded
    """

    def __init__(self, *args: ta.Any, **kwargs: ta.Any) -> None:
        super().__init__()

        self._keys: ta.MutableMapping[IdentityWeakKeyDictionary._Id[K], K] = weakref.WeakValueDictionary()
        self._values: ta.MutableMapping[IdentityWeakKeyDictionary._Id[K], V] = weakref.WeakKeyDictionary()

        for k, v in lang.yield_dict_init(*args, **kwargs):
            self[k] = v

    def __len__(self) -> int:
        return len(self._keys)

    def __iter__(self) -> ta.Iterator[K]:
        return iter(self._keys.values())

    class _Id(ta.Generic[T]):
        def __init__(self, key: T) -> None:
            super().__init__()

            self._id = id(key)
            self._key_ref = weakref.ref(key)

        def __repr__(self) -> str:
            return f'{self.__class__.__name__}<id={self._id}>'

        def __hash__(self) -> int:
            return self._id

        def __eq__(self, other: object) -> bool:
            return (
                type(other) is type(self) and
                self._id == other._id and  # type: ignore  # noqa
                self._key_ref() is other._key_ref()  # type: ignore  # noqa
            )

        def __ne__(self, other: object) -> bool:
            return not (self == other)

    def __getitem__(self, key: K) -> V:
        return self._values.__getitem__(self._Id(key))

    def __setitem__(self, key: K, value: V) -> None:
        id_obj = self._Id(key)
        self._keys.__setitem__(id_obj, key)
        self._values.__setitem__(id_obj, value)

    def __delitem__(self, key: K) -> None:
        self._values.__delitem__(self._Id(key))
        self._keys.__delitem__(self._Id(key))


class IdentityWeakSet(ta.MutableSet[T]):
    def __init__(self, init: ta.Iterable[T] | None = None) -> None:
        super().__init__()

        self._dict: IdentityWeakKeyDictionary[T, None] = IdentityWeakKeyDictionary()

        if init is not None:
            for e in init:
                self._dict[e] = None

    def add(self, value):
        self._dict[value] = None

    def discard(self, value):
        del self._dict[value]

    def __contains__(self, x):
        return x in self._dict

    def __len__(self):
        return len(self._dict)

    def __iter__(self):
        return self._dict.keys()



import enum


##


class ProcessorPriority(enum.IntEnum):
    BOOTSTRAP = enum.auto()

    PRE_GENERATION = enum.auto()
    GENERATION = enum.auto()
    POST_GENERATION = enum.auto()

    PRE_SLOTS = enum.auto()
    SLOTS = enum.auto()
    POST_SLOTS = enum.auto()



import collections.abc

from ... import check
from ... import lang
from ... import reflect as rfl
from ...funcs import match as mfs
from ..base.contexts import MarshalContext
from ..base.contexts import UnmarshalContext
from ..base.types import Marshaler
from ..base.types import Unmarshaler
from ..factories.match import MarshalerFactoryMatchClass
from ..factories.match import UnmarshalerFactoryMatchClass
from .iterables import DEFAULT_ITERABLE_CONCRETE_TYPES
from .iterables import IterableMarshaler
from .iterables import IterableUnmarshaler


##


class SequenceNotStrMarshalerFactory(MarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: isinstance(rty, rfl.Protocol) and rty.cls is lang.SequenceNotStr)
    def _build_generic(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        gty = check.isinstance(rty, rfl.Protocol)
        return IterableMarshaler(ctx.make(check.single(gty.args)))


class SequenceNotStrUnmarshalerFactory(UnmarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: isinstance(rty, rfl.Protocol) and rty.cls is lang.SequenceNotStr)
    def _build_generic(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        gty = check.isinstance(rty, rfl.Protocol)
        cty = DEFAULT_ITERABLE_CONCRETE_TYPES[collections.abc.Sequence]  # type: ignore[type-abstract]
        return IterableUnmarshaler(cty, ctx.make(check.single(gty.args)))



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import contextlib
import logging
import os
import subprocess
import time
import typing as ta

from ..lite.abstract import Abstract
from ..lite.timeouts import Timeout
from ..logs.protocols import LoggerLike
from .wrap import subprocess_maybe_shell_wrap_exec


T = ta.TypeVar('T')
SubprocessChannelOption = ta.Literal['pipe', 'stdout', 'devnull']  # ta.TypeAlias


##


# Valid channel type kwarg values:
#  - A special flag negative int
#  - A positive fd int
#  - A file-like object
#  - None

SUBPROCESS_CHANNEL_OPTION_VALUES: ta.Mapping[SubprocessChannelOption, int] = {
    'pipe': subprocess.PIPE,
    'stdout': subprocess.STDOUT,
    'devnull': subprocess.DEVNULL,
}


##


class VerboseCalledProcessError(subprocess.CalledProcessError):
    @classmethod
    def from_std(cls, e: subprocess.CalledProcessError) -> 'VerboseCalledProcessError':
        return cls(
            e.returncode,
            e.cmd,
            output=e.output,
            stderr=e.stderr,
        )

    def __str__(self) -> str:
        msg = super().__str__()
        if self.output is not None:
            msg += f' Output: {self.output!r}'
        if self.stderr is not None:
            msg += f' Stderr: {self.stderr!r}'
        return msg


class BaseSubprocesses(Abstract):
    DEFAULT_LOGGER: ta.ClassVar[ta.Optional[LoggerLike]] = None

    def __init__(
            self,
            *,
            log: ta.Optional[LoggerLike] = None,
            try_exceptions: ta.Optional[ta.Tuple[ta.Type[Exception], ...]] = None,
    ) -> None:
        super().__init__()

        self._log = log if log is not None else self.DEFAULT_LOGGER
        self._try_exceptions = try_exceptions if try_exceptions is not None else self.DEFAULT_TRY_EXCEPTIONS

    def set_logger(self, log: ta.Optional[LoggerLike]) -> None:
        self._log = log

    #

    def prepare_args(
            self,
            *cmd: str,
            env: ta.Optional[ta.Mapping[str, ta.Any]] = None,
            extra_env: ta.Optional[ta.Mapping[str, ta.Any]] = None,
            quiet: bool = False,
            shell: bool = False,
            **kwargs: ta.Any,
    ) -> ta.Tuple[ta.Tuple[ta.Any, ...], ta.Dict[str, ta.Any]]:
        if self._log:
            self._log.debug('Subprocesses.prepare_args: cmd=%r', cmd)
            if extra_env:
                self._log.debug('Subprocesses.prepare_args: extra_env=%r', extra_env)

        #

        if extra_env:
            env = {**(env if env is not None else os.environ), **extra_env}

        #

        if quiet and 'stderr' not in kwargs:
            if self._log and not self._log.isEnabledFor(logging.DEBUG):
                kwargs['stderr'] = subprocess.DEVNULL

        for chk in ('stdout', 'stderr'):
            try:
                chv = kwargs[chk]
            except KeyError:
                continue
            kwargs[chk] = SUBPROCESS_CHANNEL_OPTION_VALUES.get(chv, chv)

        #

        if not shell:
            cmd = subprocess_maybe_shell_wrap_exec(*cmd)

        #

        if 'timeout' in kwargs:
            kwargs['timeout'] = Timeout.of(kwargs['timeout']).or_(None)

        #

        return cmd, dict(
            env=env,
            shell=shell,
            **kwargs,
        )

    @contextlib.contextmanager
    def wrap_call(
            self,
            *cmd: ta.Any,
            raise_verbose: bool = False,
            **kwargs: ta.Any,
    ) -> ta.Iterator[None]:
        start_time = time.time()
        try:
            if self._log:
                self._log.debug('Subprocesses.wrap_call.try: cmd=%r', cmd)

            yield

        except Exception as exc:  # noqa
            if self._log:
                self._log.debug('Subprocesses.wrap_call.except: exc=%r', exc)

            if (
                    raise_verbose and
                    isinstance(exc, subprocess.CalledProcessError) and
                    not isinstance(exc, VerboseCalledProcessError) and
                    (exc.output is not None or exc.stderr is not None)
            ):
                raise VerboseCalledProcessError.from_std(exc) from exc

            raise

        finally:
            end_time = time.time()
            elapsed_s = end_time - start_time

            if self._log:
                self._log.debug('Subprocesses.wrap_call.finally: elapsed_s=%f cmd=%r', elapsed_s, cmd)

    @contextlib.contextmanager
    def prepare_and_wrap(
            self,
            *cmd: ta.Any,
            raise_verbose: bool = False,
            **kwargs: ta.Any,
    ) -> ta.Iterator[ta.Tuple[
        ta.Tuple[ta.Any, ...],
        ta.Dict[str, ta.Any],
    ]]:
        cmd, kwargs = self.prepare_args(*cmd, **kwargs)

        with self.wrap_call(
                *cmd,
                raise_verbose=raise_verbose,
                **kwargs,
        ):
            yield cmd, kwargs

    #

    DEFAULT_TRY_EXCEPTIONS: ta.Tuple[ta.Type[Exception], ...] = (
        FileNotFoundError,
        subprocess.CalledProcessError,
    )

    def try_fn(
            self,
            fn: ta.Callable[..., T],
            *cmd: str,
            try_exceptions: ta.Optional[ta.Tuple[ta.Type[Exception], ...]] = None,
            **kwargs: ta.Any,
    ) -> ta.Union[T, Exception]:
        if try_exceptions is None:
            try_exceptions = self._try_exceptions

        try:
            return fn(*cmd, **kwargs)

        except try_exceptions as e:  # noqa
            if self._log and self._log.isEnabledFor(logging.DEBUG):
                self._log.exception('command failed')
            return e

    async def async_try_fn(
            self,
            fn: ta.Callable[..., ta.Awaitable[T]],
            *cmd: ta.Any,
            try_exceptions: ta.Optional[ta.Tuple[ta.Type[Exception], ...]] = None,
            **kwargs: ta.Any,
    ) -> ta.Union[T, Exception]:
        if try_exceptions is None:
            try_exceptions = self._try_exceptions

        try:
            return await fn(*cmd, **kwargs)

        except try_exceptions as e:  # noqa
            if self._log and self._log.isEnabledFor(logging.DEBUG):
                self._log.exception('command failed')
            return e



from ... import lang
from ... import reflect as rfl
from ...funcs import match as mfs
from ..base.contexts import MarshalContext
from ..base.contexts import UnmarshalContext
from ..base.types import Marshaler
from ..base.types import MarshalerFactory
from ..base.types import MarshalerMaker
from ..base.types import Unmarshaler
from ..base.types import UnmarshalerFactory
from ..base.types import UnmarshalerMaker


##


class MarshalerFactoryMatchClass(
    MarshalerFactory,
    mfs.MatchFnClass[[MarshalContext, rfl.Type], Marshaler],
    lang.Abstract,
):
    @property
    def make_marshaler(self) -> MarshalerMaker:
        return self  # noqa


class UnmarshalerFactoryMatchClass(
    UnmarshalerFactory,
    mfs.MatchFnClass[[UnmarshalContext, rfl.Type], Unmarshaler],
    lang.Abstract,
):
    @property
    def make_unmarshaler(self) -> UnmarshalerMaker:
        return self  # noqa



import abc
import typing as ta

from .. import lang
from .elements import Elemental
from .elements import as_elements
from .inspect import KwargsTarget
from .keys import Key


if ta.TYPE_CHECKING:
    from .impl import injector as _injector
else:
    _injector = lang.proxy_import('.impl.injector', __package__)


T = ta.TypeVar('T')


##


class AsyncInjector(lang.Abstract):
    @abc.abstractmethod
    def try_provide(self, key: ta.Any) -> ta.Awaitable[lang.Maybe[ta.Any]]:
        raise NotImplementedError

    @abc.abstractmethod
    def provide(self, key: ta.Any) -> ta.Awaitable[ta.Any]:
        raise NotImplementedError

    @abc.abstractmethod
    def provide_kwargs(self, kt: KwargsTarget) -> ta.Awaitable[ta.Mapping[str, ta.Any]]:
        raise NotImplementedError

    @abc.abstractmethod
    def inject(self, obj: ta.Any) -> ta.Awaitable[ta.Any]:
        raise NotImplementedError

    def __getitem__(
            self,
            target: Key[T] | type[T],
    ) -> ta.Awaitable[T]:
        return self.provide(target)


def create_async_injector(*args: Elemental) -> ta.Awaitable[AsyncInjector]:
    return _injector.create_async_injector(as_elements(*args))



# MIT License
# ===========
#
# Copyright (c) 2006 Bob Ippolito
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# https://github.com/simplejson/simplejson/blob/6932004966ab70ef47250a2b3152acd8c904e6b5/simplejson/encoder.py
# https://github.com/simplejson/simplejson/blob/6932004966ab70ef47250a2b3152acd8c904e6b5/simplejson/scanner.py
import json
import re
import sys
import typing as ta

from ... import lang


##


_ESCAPE_PAT = re.compile(r'[\x00-\x1f\\"]')
_ESCAPE_ASCII_PAT = re.compile(r'[\\"]|[^\ -~]')

ESCAPE_MAP: ta.Mapping[str, str] = {
    **{
        chr(i): f'\\u{i:04x}'
        for i in range(0x20)
    },
    '\\': '\\\\',
    '"': '\\"',
    '\b': '\\b',
    '\f': '\\f',
    '\n': '\\n',
    '\r': '\\r',
    '\t': '\\t',
}


def _convert_to_string(s: str | bytes) -> str:
    if isinstance(s, bytes):
        return str(s, 'utf-8')

    elif type(s) is not str:
        # Convert a str subclass instance to exact str. Raise a TypeError otherwise.
        return str.__str__(s)

    else:
        return s


def _handle_encode_quote_args(
        q: str | None,
        lq: str | None,
        rq: str | None,
) -> tuple[str, str]:
    if q is None:
        q = '"'
    if lq is None:
        lq = q
    if rq is None:
        rq = q
    return (lq, rq)


def encode_string(
        s: str | bytes,
        q: str | None = None,
        *,
        lq: str | None = None,
        rq: str | None = None,
        escape_map: ta.Mapping[str, str] | None = None,
        ensure_ascii: bool = False,
        process_chunks: ta.Callable[[list[str]], ta.Iterable[str]] | None = None,
) -> str:
    """Return a JSON representation of a Python string."""

    s = _convert_to_string(s)

    lq, rq = _handle_encode_quote_args(q, lq, rq)

    if escape_map is None:
        escape_map = ESCAPE_MAP

    if ensure_ascii:
        pat = _ESCAPE_ASCII_PAT

        def replace(c):
            try:
                return escape_map[c]

            except KeyError:
                n = ord(c)
                if n < 0x10000:
                    return f'\\u{n:04x}'

                # surrogate pair
                n -= 0x10000
                s1 = 0xD800 | ((n >> 10) & 0x3FF)
                s2 = 0xDC00 | (n & 0x3FF)
                return f'\\u{s1:04x}\\u{s2:04x}'

    else:
        pat = _ESCAPE_PAT

        def replace(c):
            return escape_map[c]

    if process_chunks is not None:
        chunks: list[str] = [
            replace(p.group(0)) if isinstance(p, re.Match) else p
            for p in lang.iter_pat(pat, s)
        ]

        return ''.join([
            lq,
            *process_chunks(chunks),
            rq,
        ])

    else:
        return ''.join([
            lq,
            pat.sub(lambda m: replace(m.group(0)), s),
            rq,
        ])


##


_FOUR_DIGIT_HEX_PAT = re.compile(r'^[0-9a-fA-F]{4}$')


def _scan_four_digit_hex(
        s: str,
        end: int,
):
    """Scan a four digit hex number from s[end:end + 4]."""

    msg = 'Invalid \\uXXXX escape sequence'
    esc = s[end: end + 4]
    if not _FOUR_DIGIT_HEX_PAT.match(esc):
        raise json.JSONDecodeError(msg, s, end - 2)

    try:
        return int(esc, 16), end + 4
    except ValueError:
        raise json.JSONDecodeError(msg, s, end - 2) from None


_STRING_CHUNK_PAT = re.compile(r'(.*?)(["\\\x00-\x1f])', re.VERBOSE | re.MULTILINE | re.DOTALL)

BACKSLASH_MAP: ta.Mapping[str, str] = {
    '"': '"',
    '\\': '\\',
    '/': '/',
    'b': '\b',
    'f': '\f',
    'n': '\n',
    'r': '\r',
    't': '\t',
}


def parse_string(
        s: str,
        idx: int = 0,
        *,
        strict: bool = True,
) -> tuple[str, int]:
    """
    Scan the string s for a JSON string. Idx is the index of the quote that starts the JSON string. Unescapes all valid
    JSON string escape sequences and raises ValueError on attempt to decode an invalid string. If strict is False then
    literal control characters are allowed in the string.

    Returns a tuple of the decoded string and the index of the character in s after the end quote.
    """

    if s[idx] != '"':
        raise json.JSONDecodeError('No opening string quotes at', s, idx)

    chunks: list[str] = []
    end = idx + 1
    while True:
        chunk = _STRING_CHUNK_PAT.match(s, end)
        if chunk is None:
            raise json.JSONDecodeError('Unterminated string starting at', s, idx)

        prev_end = end
        end = chunk.end()
        content, terminator = chunk.groups()
        # Content contains zero or more unescaped string characters
        if content:
            chunks.append(content)

        # Terminator is the end of string, a literal control character, or a backslash denoting that an escape sequence
        # follows
        if terminator == '"':
            break
        elif terminator != '\\':
            if strict:
                msg = 'Invalid control character %r at'
                raise json.JSONDecodeError(msg, s, prev_end)
            else:
                chunks.append(terminator)
                continue

        try:
            esc = s[end]
        except IndexError:
            raise json.JSONDecodeError('Unterminated string starting at', s, idx) from None

        # If not a unicode escape sequence, must be in the lookup table
        if esc != 'u':
            try:
                char = BACKSLASH_MAP[esc]
            except KeyError:
                msg = 'Invalid \\X escape sequence %r'
                raise json.JSONDecodeError(msg, s, end) from None
            end += 1

        else:
            # Unicode escape sequence
            uni, end = _scan_four_digit_hex(s, end + 1)

            # Check for surrogate pair on UCS-4 systems Note that this will join high/low surrogate pairs but will also
            # pass unpaired surrogates through
            if (
                    sys.maxunicode > 65535 and
                    uni & 0xFC00 == 0xD800 and
                    s[end: end + 2] == '\\u'
            ):
                uni2, end2 = _scan_four_digit_hex(s, end + 2)
                if uni2 & 0xFC00 == 0xDC00:
                    uni = 0x10000 + (((uni - 0xD800) << 10) | (uni2 - 0xDC00))
                    end = end2

            char = chr(uni)

        # Append the unescaped character
        chunks.append(char)

    return ''.join(chunks), end


def try_parse_string(
        s: str,
        idx: int = 0,
        *,
        strict: bool = True,
) -> tuple[str, int] | None:
    if not s or s[idx] != '"':
        return None

    return parse_string(s, idx, strict=strict)


##


_NUMBER_PAT = re.compile(
    r'(-?(?:0|[1-9]\d*))(\.\d+)?([eE][-+]?\d+)?',
    (re.VERBOSE | re.MULTILINE | re.DOTALL),
)


def try_parse_number(
        s: str,
        idx: int = 0,
        *,
        parse_float: ta.Callable[[str], float] = float,
        parse_int: ta.Callable[[str], int] = int,
) -> tuple[int | float, int] | None:
    if (m := _NUMBER_PAT.match(s, idx)) is None:
        return None

    integer, frac, exp = m.groups()
    if frac or exp:
        res = parse_float(integer + (frac or '') + (exp or ''))
    else:
        res = parse_int(integer)
    return res, m.end()



import typing as ta

from ... import lang
from .registries import Registry
from .registries import RegistryItem


##


class Config(RegistryItem, lang.Abstract):
    pass


ConfigRegistry: ta.TypeAlias = Registry[Config]


EMPTY_CONFIG_REGISTRY = ConfigRegistry().seal()



"""
Insufficient alt impls:
 - `sys.meta_path`:
  - need access to the `fromlist` and `level` arguments passed to `__import__`
  - need to return fake modules to the import operation which are not added to `sys.modules`
 - `sys.addaudithook`: cannot prevent import or inject result
 - `sys.settrace` / bytecode tracing: same
 - jit bytecode rewriting: slower than just importing everything

Possible alt impls:
 - aot static analysis, codegen, compare, if valid skip ctxmgr body and inject proxies, otherwise warn and run
"""
import builtins
import contextlib
import functools
import sys
import threading
import types
import typing as ta


##


class ImportCaptureError(Exception):
    pass


class ImportCaptureErrors:
    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError

    class HookError(ImportCaptureError):
        pass

    class AttrError(ImportCaptureError):
        def __init__(self, module: str | None, name: str) -> None:
            super().__init__()

            self.module = module
            self.name = name

        def __repr__(self) -> str:
            return f'{self.__class__.__qualname__}(module={self.module!r}, name={self.name!r})'

    class ImportError(ImportCaptureError):  # noqa
        def __init__(self, module: str, from_list: ta.Sequence[str] | None) -> None:
            super().__init__()

            self.module = module
            self.from_list = from_list

        def __repr__(self) -> str:
            return f'{self.__class__.__qualname__}(module={self.module!r}, from_list={self.from_list!r})'

    class ImportStarForbiddenError(ImportError):
        pass

    class UncapturedImportForbiddenError(ImportError):
        pass

    class UnreferencedImportsError(ImportCaptureError):
        def __init__(self, unreferenced: ta.Mapping[str, ta.Sequence[str | None]]) -> None:
            super().__init__()

            self.unreferenced = unreferenced

        def __repr__(self) -> str:
            return f'{self.__class__.__qualname__}(unreferenced={self.unreferenced!r})'

    class CaptureInProgressError(ImportCaptureError):
        pass


##


class _ImportCaptureHook:
    class ModuleSpec(ta.NamedTuple):
        name: str
        level: int

        def __str__(self) -> str:
            return f'{"." * self.level}{self.name}'

        def __repr__(self) -> str:
            return repr(str(self))

    def __init__(
            self,
            *,
            forbid_uncaptured_imports: bool = False,
    ) -> None:
        super().__init__()

        self._forbid_uncaptured_imports = forbid_uncaptured_imports

        self._modules_by_spec: dict[_ImportCaptureHook.ModuleSpec, _ImportCaptureHook._Module] = {}
        self._modules_by_module_obj: dict[types.ModuleType, _ImportCaptureHook._Module] = {}

        self._attrs: dict[_ImportCaptureHook._ModuleAttr, tuple[_ImportCaptureHook._Module, str]] = {}

    #

    class _ModuleAttr:
        def __init__(
                self,
                module: '_ImportCaptureHook._Module',
                name: str,
        ) -> None:
            super().__init__()

            self.__module = module
            self.__name = name

        def __repr__(self) -> str:
            return f'<{self.__class__.__name__}: {f"{self.__module.spec}:{self.__name}"!r}>'

    class _Module:
        def __init__(
                self,
                spec: '_ImportCaptureHook.ModuleSpec',
                *,
                getattr_handler: ta.Callable[['_ImportCaptureHook._Module', str], ta.Any] | None = None,
        ) -> None:
            super().__init__()

            self.spec = spec

            self.module_obj = types.ModuleType(f'<{self.__class__.__qualname__}: {spec!r}>')
            if getattr_handler is not None:
                self.module_obj.__getattr__ = functools.partial(getattr_handler, self)  # type: ignore[method-assign]  # noqa
            self.initial_module_dict = dict(self.module_obj.__dict__)

            self.contents: dict[str, _ImportCaptureHook._ModuleAttr | types.ModuleType] = {}
            self.imported_whole = False

        def __repr__(self) -> str:
            return f'{self.__class__.__name__}({self.spec!r})'

    def _get_or_make_module(self, spec: ModuleSpec) -> _Module:
        try:
            return self._modules_by_spec[spec]
        except KeyError:
            pass

        module = self._Module(
            spec,
            getattr_handler=self._handle_module_getattr,
        )
        self._modules_by_spec[spec] = module
        self._modules_by_module_obj[module.module_obj] = module
        return module

    def _handle_module_getattr(self, module: _Module, attr: str) -> ta.Any:
        if attr in module.contents:
            raise ImportCaptureErrors.AttrError(str(module.spec), attr)

        v: _ImportCaptureHook._ModuleAttr | types.ModuleType
        if not module.spec.name:
            if not module.spec.level:
                raise ImportCaptureError
            cs = _ImportCaptureHook.ModuleSpec(attr, module.spec.level)
            cm = self._get_or_make_module(cs)
            cm.imported_whole = True
            v = cm.module_obj

        else:
            ma = _ImportCaptureHook._ModuleAttr(module, attr)
            self._attrs[ma] = (module, attr)
            v = ma

        module.contents[attr] = v
        setattr(module.module_obj, attr, v)
        return v

    def _handle_import(
            self,
            module: _Module,
            *,
            from_list: ta.Sequence[str] | None,
    ) -> None:
        if from_list is None:
            if module.spec.level or not module.spec.name:
                raise ImportCaptureError

            module.imported_whole = True

        else:
            for attr in from_list:
                if attr == '*':
                    raise ImportCaptureErrors.ImportStarForbiddenError(str(module.spec), from_list)

                x = getattr(module.module_obj, attr)

                bad = False
                if x is not module.contents.get(attr):
                    bad = True
                if isinstance(x, _ImportCaptureHook._ModuleAttr):
                    if self._attrs[x] != (module, attr):
                        bad = True
                elif isinstance(x, types.ModuleType):
                    if x not in self._modules_by_module_obj:
                        bad = True
                else:
                    bad = True
                if bad:
                    raise ImportCaptureErrors.AttrError(str(module.spec), attr)

    #

    _MOD_SELF_ATTR: ta.ClassVar[str] = '__import_capture__'

    def _intercept_import(
            self,
            name: str,
            *,
            globals: ta.Mapping[str, ta.Any] | None = None,  # noqa
            from_list: ta.Sequence[str] | None = None,
            level: int = 0,
    ) -> types.ModuleType | None:
        if not (
                globals is not None and
                globals.get(self._MOD_SELF_ATTR) is self
        ):
            return None

        spec = _ImportCaptureHook.ModuleSpec(name, level)
        module = self._get_or_make_module(spec)

        self._handle_import(
            module,
            from_list=from_list,
        )

        return module.module_obj

    # @abc.abstractmethod
    def hook_context(
            self,
            mod_globals: ta.MutableMapping[str, ta.Any],  # noqa
    ) -> ta.ContextManager[None]:
        raise NotImplementedError

    #

    def verify_state(
            self,
            mod_globals: ta.MutableMapping[str, ta.Any],  # noqa
    ) -> None:
        for m in self._modules_by_spec.values():
            for a, o in m.module_obj.__dict__.items():
                try:
                    i = m.initial_module_dict[a]

                except KeyError:
                    if o is not m.contents[a]:
                        raise ImportCaptureErrors.AttrError(str(m.spec), a) from None

                else:
                    if o != i:
                        raise ImportCaptureErrors.AttrError(str(m.spec), a)

    #

    def build_captured(
            self,
            mod_globals: ta.MutableMapping[str, ta.Any],  # noqa
            *,
            collect_unreferenced: bool = False,
    ) -> 'ImportCapture.Captured':
        dct: dict[_ImportCaptureHook._Module, list[tuple[str | None, str]]] = {}

        rem_whole_mods: set[_ImportCaptureHook._Module] = set()
        rem_mod_attrs: set[_ImportCaptureHook._ModuleAttr] = set()
        if collect_unreferenced:
            rem_whole_mods.update([m for m in self._modules_by_spec.values() if m.imported_whole])
            rem_mod_attrs.update(self._attrs)

        for attr, obj in mod_globals.items():
            if isinstance(obj, _ImportCaptureHook._ModuleAttr):
                try:
                    m, a = self._attrs[obj]
                except KeyError:
                    raise ImportCaptureErrors.AttrError(None, attr) from None
                dct.setdefault(m, []).append((a, attr))
                rem_mod_attrs.discard(obj)

            elif isinstance(obj, _ImportCaptureHook._Module):
                raise ImportCaptureErrors.AttrError(None, attr) from None

            elif isinstance(obj, types.ModuleType):
                try:
                    m = self._modules_by_module_obj[obj]
                except KeyError:
                    continue
                if not m.imported_whole:
                    raise RuntimeError(f'ImportCapture module {m.spec!r} not imported_whole')
                dct.setdefault(m, []).append((None, attr))
                rem_whole_mods.discard(m)

        lst: list[ImportCapture.Import] = []
        for m, ts in dct.items():
            if not m.spec.name:
                if not m.spec.level:
                    raise ImportCaptureError
                for imp_attr, as_attr in ts:
                    if not imp_attr:
                        raise RuntimeError
                    lst.append(ImportCapture.Import(
                        '.' * m.spec.level + imp_attr,
                        [(None, as_attr)],
                    ))

            else:
                lst.append(ImportCapture.Import(
                    str(m.spec),
                    ts,
                ))

        unreferenced: dict[str, list[str | None]] | None = None
        if collect_unreferenced and (rem_whole_mods or rem_mod_attrs):
            unreferenced = {}
            for m in rem_whole_mods:
                unreferenced.setdefault(str(m.spec), []).append(None)
            for ma in rem_mod_attrs:
                m, a = self._attrs[ma]
                unreferenced.setdefault(str(m.spec), []).append(a)

        return ImportCapture.Captured(
            lst,
            unreferenced,
        )


class _AbstractBuiltinImportCaptureHook(_ImportCaptureHook):
    def _new_import(
            self,
            old_import,
            name,
            globals=None,  # noqa
            locals=None,  # noqa
            fromlist=None,
            level=0,
    ):
        if (im := self._intercept_import(
                name,
                globals=globals,
                from_list=fromlist,
                level=level,
        )) is not None:
            return im

        if self._forbid_uncaptured_imports:
            raise ImportCaptureErrors.UncapturedImportForbiddenError(
                str(_ImportCaptureHook.ModuleSpec(name, level)),
                fromlist,
            )

        return old_import(
            name,
            globals=globals,
            locals=locals,
            fromlist=fromlist,
            level=level,
        )


class _UnsafeGlobalBuiltinImportCaptureHook(_AbstractBuiltinImportCaptureHook):
    @contextlib.contextmanager
    def hook_context(
            self,
            mod_globals: ta.MutableMapping[str, ta.Any],  # noqa
    ) -> ta.Iterator[None]:
        if self._MOD_SELF_ATTR in mod_globals:
            raise ImportCaptureErrors.HookError

        #

        old_import = builtins.__import__
        new_import = functools.partial(self._new_import, old_import)

        mod_globals[self._MOD_SELF_ATTR] = self
        builtins.__import__ = new_import

        try:
            yield

        finally:
            if not (
                    mod_globals[self._MOD_SELF_ATTR] is self and
                    builtins.__import__ is new_import
            ):
                raise ImportCaptureErrors.HookError

            del mod_globals[self._MOD_SELF_ATTR]
            builtins.__import__ = old_import


class _SomewhatThreadSafeGlobalBuiltinImportCaptureHook(_AbstractBuiltinImportCaptureHook):
    class _AlreadyPatchedError(Exception):
        pass

    @ta.final
    class _Patch:
        __lock: ta.ClassVar[threading.Lock] = threading.Lock()

        def __init__(self, old_import):
            self.__old_import = old_import
            self.__hooks = {}
            self.__uninstalled = False

        @classmethod
        def _add_hook(cls, mod_globals, new_import) -> '_SomewhatThreadSafeGlobalBuiltinImportCaptureHook._Patch':
            gi = id(mod_globals)
            for _ in range(1_000):
                try:
                    with cls.__lock:
                        x: ta.Any = builtins.__import__
                        p: _SomewhatThreadSafeGlobalBuiltinImportCaptureHook._Patch
                        if x.__class__ is cls:
                            p = x
                            if p.__uninstalled:  # noqa
                                raise _SomewhatThreadSafeGlobalBuiltinImportCaptureHook._AlreadyPatchedError  # noqa
                        else:
                            p = cls(x)
                            builtins.__import__ = p
                        p.__hooks[gi] = (mod_globals, new_import)
                        return p
                except _SomewhatThreadSafeGlobalBuiltinImportCaptureHook._AlreadyPatchedError:
                    pass
            raise ImportCaptureErrors.HookError('Failed to install builtins hook')

        def _remove_hook(self, mod_globals, *, no_raise=False):
            gi = id(mod_globals)
            with self.__lock:
                tg, _ = self.__hooks[gi]
                del self.__hooks[gi]
                if not self.__uninstalled and not self.__hooks:
                    self.__uninstalled = True
                    if builtins.__import__ is not self:
                        if not no_raise:
                            # TODO: warn?
                            raise ImportCaptureErrors.HookError('Unexpected builtins hook')
                    else:
                        builtins.__import__ = self.__old_import
                if tg is not mod_globals:
                    if not no_raise:
                        # TODO: warn?
                        raise ImportCaptureErrors.HookError('Mismatched globals')

        def __call__(
                self,
                name,
                globals=None,  # noqa
                locals=None,  # noqa
                fromlist=None,
                level=0,
        ):
            if globals is not None and (tup := self.__hooks.get(id(globals))) is not None:
                tg, tf = tup
                if tg is globals:
                    return tf(
                        self.__old_import,
                        name,
                        globals=globals,
                        locals=locals,
                        fromlist=fromlist,
                        level=level,
                    )
                else:
                    self._remove_hook(tg, no_raise=True)

            return self.__old_import(
                name,
                globals=globals,
                locals=locals,
                fromlist=fromlist,
                level=level,
            )

    @contextlib.contextmanager
    def hook_context(
            self,
            mod_globals: ta.MutableMapping[str, ta.Any],  # noqa
    ) -> ta.Iterator[None]:
        if self._MOD_SELF_ATTR in mod_globals:
            raise ImportCaptureErrors.HookError

        #

        mod_globals[self._MOD_SELF_ATTR] = self
        patch = _SomewhatThreadSafeGlobalBuiltinImportCaptureHook._Patch._add_hook(mod_globals, self._new_import)  # noqa

        try:
            yield

        finally:
            patch._remove_hook(mod_globals)  # noqa

            if mod_globals[self._MOD_SELF_ATTR] is not self:
                raise ImportCaptureErrors.HookError
            del mod_globals[self._MOD_SELF_ATTR]


class _FrameBuiltinImportCaptureHook(_AbstractBuiltinImportCaptureHook):
    def __init__(
            self,
            *,
            _frame: types.FrameType,
            **kwargs: ta.Any,
    ) -> None:
        super().__init__(**kwargs)

        self._frame = _frame

    @classmethod
    def _set_frame_builtins(
            cls,
            frame: types.FrameType,
            new_builtins: dict[str, ta.Any],
    ) -> None:
        # FIXME: cext...
        frame.f_builtins = new_builtins  # type: ignore[misc]  # noqa

    @contextlib.contextmanager
    def hook_context(
            self,
            mod_globals: ta.MutableMapping[str, ta.Any],  # noqa
    ) -> ta.Iterator[None]:
        if self._MOD_SELF_ATTR in mod_globals:
            raise ImportCaptureErrors.HookError

        #

        old_builtins = self._frame.f_builtins
        old_import = old_builtins['__import__']
        new_import = functools.partial(self._new_import, old_import)

        mod_globals[self._MOD_SELF_ATTR] = self
        new_builtins = dict(old_builtins)
        new_builtins['__import__'] = new_import
        self._set_frame_builtins(self._frame, new_builtins)

        try:
            yield

        finally:
            if not (
                    mod_globals[self._MOD_SELF_ATTR] is self and
                    self._frame.f_builtins is new_builtins
            ):
                raise ImportCaptureErrors.HookError

            del mod_globals[self._MOD_SELF_ATTR]
            self._set_frame_builtins(self._frame, old_builtins)


def _new_import_capture_hook(
        mod_globals: ta.MutableMapping[str, ta.Any],  # noqa
        *,
        stack_offset: int = 0,
        **kwargs: ta.Any,
) -> '_ImportCaptureHook':
    frame: types.FrameType | None = sys._getframe(1 + stack_offset)  # noqa
    if frame is None or frame.f_globals is not mod_globals:
        raise ImportCaptureError("Can't find importing frame")

    # return _FrameBuiltinImportCaptureHook(
    #     _frame=frame,
    #     **kwargs,
    # )

    # return _UnsafeGlobalBuiltinImportCaptureHook(**kwargs)

    return _SomewhatThreadSafeGlobalBuiltinImportCaptureHook(**kwargs)


##


class ImportCapture:
    class Import(ta.NamedTuple):
        spec: str
        attrs: ta.Sequence[tuple[str | None, str]]

    class Captured(ta.NamedTuple):
        imports: ta.Sequence['ImportCapture.Import']
        unreferenced: ta.Mapping[str, ta.Sequence[str | None]] | None

        @property
        def attrs(self) -> ta.Iterator[str]:
            for pi in self.imports:
                for _, a in pi.attrs:
                    yield a

    #

    def __init__(
            self,
            mod_globals: ta.MutableMapping[str, ta.Any],
            *,
            _hook: _ImportCaptureHook,

            disable: bool = False,
    ) -> None:
        super().__init__()

        self._mod_globals = mod_globals
        self._hook = _hook

        self._disabled = disable

    @property
    def disabled(self) -> bool:
        return self._disabled

    #

    class _Result(ta.NamedTuple):
        captured: 'ImportCapture.Captured'

    _result_: _Result | None = None

    @property
    def _result(self) -> _Result:
        if (rs := self._result_) is None:
            raise ImportCaptureErrors.CaptureInProgressError
        return rs

    @property
    def is_complete(self) -> bool:
        return self._result_ is not None

    @property
    def captured(self) -> Captured:
        return self._result.captured

    #

    @contextlib.contextmanager
    def capture(
            self,
            *,
            unreferenced_callback: ta.Callable[[ta.Mapping[str, ta.Sequence[str | None]]], None] | None = None,
            raise_unreferenced: bool = False,
    ) -> ta.Iterator[ta.Self]:
        if self._result_ is not None:
            raise ImportCaptureError('capture already complete')

        if self._disabled:
            self._result_ = ImportCapture._Result(
                ImportCapture.Captured(
                    [],
                    None,
                ),
            )
            yield self
            return

        with self._hook.hook_context(self._mod_globals):
            yield self

        self._hook.verify_state(self._mod_globals)

        blt = self._hook.build_captured(
            self._mod_globals,
            collect_unreferenced=unreferenced_callback is not None or raise_unreferenced,
        )

        if blt.unreferenced:
            if unreferenced_callback:
                unreferenced_callback(blt.unreferenced)
            if raise_unreferenced:
                raise ImportCaptureErrors.UnreferencedImportsError(blt.unreferenced)

        for pi in blt.imports:
            for _, a in pi.attrs:
                del self._mod_globals[a]

        self._result_ = ImportCapture._Result(
            blt,
        )

    #

    def update_exports(self) -> None:
        cap = self._result.captured

        try:
            al: ta.Any = self._mod_globals['__all__']
        except KeyError:
            al = self._mod_globals['__all__'] = [k for k in self._mod_globals if not k.startswith('_')]
        else:
            if not isinstance(al, ta.MutableSequence):
                al = self._mod_globals['__all__'] = list(al)

        al_s = set(al)
        for a in cap.attrs:
            if a not in al_s:
                al.append(a)
                al_s.add(a)



import dataclasses as dc
import typing as ta

from ... import check
from ... import reflect as rfl
from ..base.contexts import MarshalContext
from ..base.contexts import UnmarshalContext
from ..base.types import Marshaler
from ..base.types import Unmarshaler
from ..base.values import Value
from ..factories.simple import SimpleMarshalerFactory
from ..factories.simple import SimpleUnmarshalerFactory


##


@dc.dataclass(frozen=True)
class OptionalMarshaler(Marshaler):
    e: Marshaler

    def marshal(self, ctx: MarshalContext, o: ta.Any | None) -> Value:
        if o is None:
            return None
        return self.e.marshal(ctx, o)


class OptionalMarshalerFactory(SimpleMarshalerFactory):
    def guard(self, ctx: MarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, rfl.Union) and rty.is_optional

    def fn(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        return OptionalMarshaler(ctx.make(check.isinstance(rty, rfl.Union).without_none()))


@dc.dataclass(frozen=True)
class OptionalUnmarshaler(Unmarshaler):
    e: Unmarshaler

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any | None:
        if v is None:
            return None
        return self.e.unmarshal(ctx, v)


class OptionalUnmarshalerFactory(SimpleUnmarshalerFactory):
    def guard(self, ctx: UnmarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, rfl.Union) and rty.is_optional

    def fn(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        return OptionalUnmarshaler(ctx.make(check.isinstance(rty, rfl.Union).without_none()))



from .flavors import (  # noqa
    ContextManagerAdapter,
    Flavor,
    adapt,
    adapt_context,
    from_anyio,
    from_anyio_context,
    from_asyncio,
    from_asyncio_context,
    from_trio,
    from_trio_context,
    get_flavor,
    mark_anyio,
    mark_asyncio,
    mark_flavor,
    mark_trio,
    with_adapter_loop,
)



import abc
import typing as ta

from .. import dataclasses as dc
from .. import lang
from .impl.origins import HasOriginsImpl


##


class Element(HasOriginsImpl, lang.Abstract, lang.PackageSealed):
    """Note: inheritors must be dataclasses."""


class ElementGenerator(lang.Abstract, lang.PackageSealed):
    @abc.abstractmethod
    def __iter__(self) -> ta.Iterator[Element]:
        raise NotImplementedError


@dc.dataclass(frozen=True)
class Elements(lang.Final):
    es: ta.Collection[Element] | None = None
    cs: ta.Collection['Elements'] | None = None

    def __iter__(self) -> ta.Generator[Element]:
        if self.es:
            yield from self.es
        if self.cs:
            for c in self.cs:
                yield from c


Elemental: ta.TypeAlias = ta.Union[  # noqa
    Element,
    Elements,
    ElementGenerator,
]


def as_elements(*args: Elemental) -> Elements:
    es: list[Element] = []
    cs: list[Elements] = []

    def rec(a):
        if isinstance(a, Element):
            es.append(a)
        elif isinstance(a, Elements):
            cs.append(a)
        elif isinstance(a, ElementGenerator):
            for n in a:
                rec(n)
        else:
            raise TypeError(a)

    for a in args:
        rec(a)

    if not es and len(cs) == 1:
        return next(iter(cs))

    return Elements(
        es if es else None,
        cs if cs else None,
    )



import typing as ta

from ... import reflect as rfl
from ...funcs import match as mfs
from ..base.contexts import MarshalContext
from ..base.contexts import UnmarshalContext
from ..base.types import Marshaler
from ..base.types import MarshalerFactory
from ..base.types import MarshalerMaker
from ..base.types import Unmarshaler
from ..base.types import UnmarshalerFactory
from ..base.types import UnmarshalerMaker


##


class MultiMarshalerFactory(MarshalerFactory):
    def __init__(
            self,
            fs: ta.Iterable[MarshalerFactory],
            *,
            strict: bool = False,
    ) -> None:
        super().__init__()

        self._fs = list(fs)
        self._mmf: mfs.MultiMatchFn[[MarshalContext, rfl.Type], Marshaler] = mfs.MultiMatchFn(
            [f.make_marshaler for f in self._fs],
            strict=strict,
        )

    @property
    def make_marshaler(self) -> MarshalerMaker:
        return self._mmf


class MultiUnmarshalerFactory(UnmarshalerFactory):
    def __init__(
            self,
            fs: ta.Iterable[UnmarshalerFactory],
            *,
            strict: bool = False,
    ) -> None:
        super().__init__()

        self._fs = list(fs)
        self._mmf: mfs.MultiMatchFn[[UnmarshalContext, rfl.Type], Unmarshaler] = mfs.MultiMatchFn(
            [f.make_unmarshaler for f in self._fs],
            strict=strict,
        )

    @property
    def make_unmarshaler(self) -> UnmarshalerMaker:
        return self._mmf



import threading
import typing as ta

from ... import check
from ... import reflect as rfl
from ...funcs import match as mfs
from ..base.contexts import MarshalContext
from ..base.contexts import UnmarshalContext
from ..base.types import Marshaler
from ..base.types import MarshalerFactory
from ..base.types import MarshalerMaker
from ..base.types import Unmarshaler
from ..base.types import UnmarshalerFactory
from ..base.types import UnmarshalerMaker


R = ta.TypeVar('R')
C = ta.TypeVar('C')


##


class _TypeCacheFactory(mfs.MatchFn[[C, rfl.Type], R]):
    def __init__(self, f: mfs.MatchFn[[C, rfl.Type], R]) -> None:
        super().__init__()

        self._f = f
        self._dct: dict[rfl.Type, R | None] = {}
        self._lock = threading.RLock()

    def guard(self, ctx: C, rty: rfl.Type) -> bool:
        check.isinstance(rty, rfl.TYPES)

        try:
            return self._dct[rty] is not None
        except KeyError:
            pass

        with self._lock:
            try:
                e = self._dct[rty]

            except KeyError:
                if self._f.guard(ctx, rty):
                    return True
                else:
                    self._dct[rty] = None
                    return False

            else:
                return e is not None

    def fn(self, ctx: C, rty: rfl.Type) -> R:
        check.isinstance(rty, rfl.TYPES)

        try:
            e = self._dct[rty]
        except KeyError:
            pass
        else:
            if e is None:
                raise mfs.MatchGuardError(ctx, rty)
            else:
                return e

        with self._lock:
            try:
                e = self._dct[rty]

            except KeyError:
                try:
                    ret = self._f(ctx, rty)
                except mfs.MatchGuardError:
                    self._dct[rty] = None
                    raise
                else:
                    self._dct[rty] = ret
                    return ret

            if e is None:
                raise mfs.MatchGuardError(ctx, rty)
            else:
                return e


##


class TypeCacheMarshalerFactory(MarshalerFactory):
    def __init__(self, f: MarshalerFactory) -> None:
        super().__init__()

        self._f = f
        self._tcf: _TypeCacheFactory[MarshalContext, Marshaler] = _TypeCacheFactory(f.make_marshaler)

    @property
    def make_marshaler(self) -> MarshalerMaker:
        return self._tcf


class TypeCacheUnmarshalerFactory(UnmarshalerFactory):
    def __init__(self, f: UnmarshalerFactory) -> None:
        super().__init__()

        self._f = f
        self._tcf: _TypeCacheFactory[UnmarshalContext, Unmarshaler] = _TypeCacheFactory(f.make_unmarshaler)

    @property
    def make_unmarshaler(self) -> UnmarshalerMaker:
        return self._tcf



import typing as ta

from .. import check
from .. import dataclasses as dc
from .. import lang
from .accessor import TypedValuesAccessor
from .consumer import TypedValuesConsumer
from .values import TypedValue
from .values import UniqueTypedValue


TypedValueT = ta.TypeVar('TypedValueT', bound='TypedValue')
TypedValueU = ta.TypeVar('TypedValueU', bound='TypedValue')

UniqueTypedValueT = ta.TypeVar('UniqueTypedValueT', bound='UniqueTypedValue')


##


@dc.dataclass()
class DuplicateUniqueTypedValueError(Exception, ta.Generic[UniqueTypedValueT]):
    cls: type[UniqueTypedValueT]
    new: UniqueTypedValueT
    old: UniqueTypedValueT


class TypedValues(
    TypedValuesAccessor[TypedValueT],
    lang.Final,
    ta.Generic[TypedValueT],
):
    def __init__(
            self,
            *tvs: TypedValueT,
            override: bool = False,
            check_type: type | tuple[type, ...] | None = None,
    ) -> None:
        super().__init__()

        if tvs:
            tmp: list = []
            udct: dict = {}
            for tv in tvs:
                if check_type is not None:
                    check.isinstance(tv, check_type)
                if isinstance(tv, UniqueTypedValue):
                    utvc = tv._unique_typed_value_cls  # noqa
                    if not override:
                        try:
                            exu = udct[utvc]
                        except KeyError:
                            pass
                        else:
                            raise DuplicateUniqueTypedValueError(utvc, tv, check.single(exu))
                    ulst = udct.setdefault(utvc, [])
                    ulst.append(tv)
                    tmp.append((utvc, tv, ulst, len(ulst)))
                elif isinstance(tv, TypedValue):
                    tmp.append(tv)
                else:
                    raise TypeError(tv)

            lst: list = []
            dct: dict = {}
            for obj in tmp:
                if isinstance(obj, tuple):
                    utvc, tv, ulst, idx = obj
                    if idx == len(ulst):
                        lst.append(tv)
                        dct[utvc] = tv
                else:
                    tv = obj
                    lst.append(tv)
                    dct.setdefault(type(tv), []).append(tv)

            tup = tuple(lst)
            dct = {
                k: tuple(v) if isinstance(v, list) else v
                for k, v in dct.items()
            }
            dct2 = {
                **dct,
                **{type(v): v for v in dct.values() if isinstance(v, UniqueTypedValue)},
            }

        else:
            tup = ()
            dct = {}
            dct2 = {}

        self._tup: tuple[TypedValueT, ...] = tup
        self._dct: dict[type[TypedValueT], TypedValueT | tuple[TypedValueT, ...]] = dct
        self._dct2: dict[type[TypedValueT], TypedValueT | tuple[TypedValueT, ...]] = dct2

    #

    def without(self, *tys: type) -> ta.Iterator[TypedValueT]:
        for o in self._tup:
            if isinstance(o, tys):
                continue
            yield o

    #

    def update(self, *tvs, override: bool = False) -> 'TypedValues':
        return TypedValues(*self._tup, *tvs, override=override)

    def discard(self, *tys: type) -> 'TypedValues':
        return TypedValues(*self.without(*tys))

    #

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({", ".join(map(repr, self._tup))})'

    #

    _set: frozenset[TypedValueT]

    def to_set(self) -> frozenset[TypedValueT]:
        try:
            return self._set
        except AttributeError:
            pass
        s = frozenset(self._tup)
        self._set = s
        return s

    #

    _hash: int

    def __hash__(self) -> int:
        try:
            return self._hash
        except AttributeError:
            pass
        h = hash(self._tup)
        self._hash = h
        return h

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return (
            other is self or
            other._tup == self._tup
        )

    #

    def __iter__(self) -> ta.Iterator[TypedValueT]:
        return iter(self._tup)

    def __len__(self) -> int:
        return len(self._tup)

    def __bool__(self) -> bool:
        return bool(self._tup)

    #

    def keys(self) -> ta.KeysView[type[TypedValueT]]:
        return self._dct.keys()

    def values(self) -> ta.ValuesView[TypedValueT | tuple[TypedValueT, ...]]:
        return self._dct.values()

    def items(self) -> ta.ItemsView[type[TypedValueT], TypedValueT | tuple[TypedValueT, ...]]:
        return self._dct.items()

    #

    def consume(self) -> TypedValuesConsumer[TypedValueT]:
        return TypedValuesConsumer(self._dct.items())

    #

    def _typed_value_contains(self, cls):
        return cls in self._dct2

    def _typed_value_getitem(self, key):
        if isinstance(key, int):
            return self._tup[key]
        elif isinstance(key, type):
            return self._dct2[check.issubclass(key, TypedValue)]
        else:
            raise TypeError(key)

    _any_dct: dict[type | tuple[type, ...], tuple[TypedValueT, ...]]

    def _typed_value_get_any(self, cls):
        try:
            any_dct = self._any_dct
        except AttributeError:
            any_dct = {}
            self._any_dct = any_dct

        try:
            return any_dct[cls]
        except KeyError:
            pass

        ret = tuple(tv for tv in self if isinstance(tv, cls))
        any_dct[cls] = ret
        return ret


collect = TypedValues


def as_collection(src: ta.Sequence[TypedValueT]) -> TypedValues[TypedValueT]:
    if isinstance(src, TypedValues):
        return src
    else:
        return TypedValues(*src)



# ruff: noqa: UP006 UP036 UP045
import typing as ta


T = ta.TypeVar('T')


##


def deep_subclasses(cls: ta.Type[T]) -> ta.Iterator[ta.Type[T]]:
    seen = set()
    todo = list(reversed(cls.__subclasses__()))
    while todo:
        cur = todo.pop()
        if cur in seen:
            continue
        seen.add(cur)
        yield cur
        todo.extend(reversed(cur.__subclasses__()))


##


def mro_owner_dict(
        instance_cls: type,
        owner_cls: ta.Optional[type] = None,
        *,
        bottom_up_key_order: bool = False,
        sort_keys: bool = False,
) -> ta.Mapping[str, ta.Tuple[type, ta.Any]]:
    if owner_cls is None:
        owner_cls = instance_cls

    mro = instance_cls.__mro__[-2::-1]
    try:
        pos = mro.index(owner_cls)
    except ValueError:
        raise TypeError(f'Owner class {owner_cls} not in mro of instance class {instance_cls}') from None

    dct: ta.Dict[str, ta.Tuple[type, ta.Any]] = {}
    if not bottom_up_key_order:
        for cur_cls in mro[:pos + 1][::-1]:
            for k, v in cur_cls.__dict__.items():
                if k not in dct:
                    dct[k] = (cur_cls, v)

    else:
        for cur_cls in mro[:pos + 1]:
            dct.update({k: (cur_cls, v) for k, v in cur_cls.__dict__.items()})

    if sort_keys:
        dct = dict(sorted(dct.items(), key=lambda t: t[0]))

    return dct


def mro_dict(
        instance_cls: type,
        owner_cls: ta.Optional[type] = None,
        *,
        bottom_up_key_order: bool = False,
        sort_keys: bool = False,
) -> ta.Mapping[str, ta.Any]:
    return {
        k: v
        for k, (o, v) in mro_owner_dict(
            instance_cls,
            owner_cls,
            bottom_up_key_order=bottom_up_key_order,
            sort_keys=sort_keys,
        ).items()
    }


def dir_dict(o: ta.Any) -> ta.Dict[str, ta.Any]:
    return {
        a: getattr(o, a)
        for a in dir(o)
    }



import typing as ta

from ... import check
from .. import abc as dbapi_abc
from .base import Adapter
from .base import Conn
from .base import Db
from .base import Rows
from .columns import Column
from .columns import Columns
from .queries import Query
from .rows import Row


##


def build_dbapi_columns(desc: ta.Sequence[dbapi_abc.DbapiColumnDescription] | None) -> Columns:
    if desc is None:
        return Columns.empty()

    cols: list[Column] = []
    for desc_col in desc:
        dbapi_col = dbapi_abc.DbapiColumnDescription_.of(desc_col)

        cols.append(Column(
            check.non_empty_str(dbapi_col.name),
        ))

    return Columns(*cols)


class DbapiRows(Rows):
    def __init__(
            self,
            cursor: dbapi_abc.DbapiCursor,
            columns: Columns,
    ) -> None:
        super().__init__()

        self._cursor = cursor
        self._columns = columns

    def _close(self) -> None:
        self._cursor.close()

        super()._close()

    @property
    def columns(self) -> Columns:
        return self._columns

    def __next__(self) -> Row:
        self._check_entered()
        values = self._cursor.fetchone()
        if values is None:
            raise StopIteration
        return Row(self._columns, values)


class DbapiConn(Conn):
    def __init__(self, conn: dbapi_abc.DbapiConnection) -> None:
        super().__init__()

        self._conn = conn

    def _close(self) -> None:
        self._conn.close()

        super()._close()

    @property
    def adapter(self) -> Adapter:
        raise NotImplementedError

    def query(self, query: Query) -> Rows:
        self._check_entered()
        cursor = self._conn.cursor()
        try:
            cursor.execute(query.text)
            columns = build_dbapi_columns(cursor.description)
            return DbapiRows(cursor, columns)

        except Exception:  # noqa
            cursor.close()
            raise


class DbapiDb(Db):
    def __init__(
            self,
            conn_fac: ta.Callable[[], dbapi_abc.DbapiConnection],
            *,
            adapter: ta.Optional['DbapiAdapter'] = None,
    ) -> None:
        super().__init__()

        self._conn_fac = conn_fac
        if adapter is None:
            adapter = DbapiAdapter()
        self._adapter = adapter

    def connect(self) -> Conn:
        self._check_entered()
        dbapi_conn = self._conn_fac()
        return DbapiConn(dbapi_conn)

    @property
    def adapter(self) -> Adapter:
        return self._adapter


class DbapiAdapter(Adapter):
    def scan_type(self, c: Column) -> type:
        raise NotImplementedError



"""
TODO:
 - heterogeneous tuples
"""
import collections.abc
import dataclasses as dc
import functools
import typing as ta

from ... import check
from ... import reflect as rfl
from ...funcs import match as mfs
from ..base.contexts import MarshalContext
from ..base.contexts import UnmarshalContext
from ..base.types import Marshaler
from ..base.types import Unmarshaler
from ..base.values import Value
from ..factories.match import MarshalerFactoryMatchClass
from ..factories.match import UnmarshalerFactoryMatchClass


##


DEFAULT_ITERABLE_CONCRETE_TYPES: dict[type[collections.abc.Iterable], type[collections.abc.Iterable]] = {
    collections.abc.Iterable: tuple,  # type: ignore
    collections.abc.Sequence: tuple,  # type: ignore
    collections.abc.MutableSequence: list,  # type: ignore
}


@dc.dataclass(frozen=True)
class IterableMarshaler(Marshaler):
    e: Marshaler

    def marshal(self, ctx: MarshalContext, o: ta.Iterable) -> Value:
        return list(map(functools.partial(self.e.marshal, ctx), o))


class IterableMarshalerFactory(MarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: isinstance(rty, rfl.Generic) and issubclass(rty.cls, collections.abc.Iterable))
    def _build_generic(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        gty = check.isinstance(rty, rfl.Generic)
        return IterableMarshaler(ctx.make(check.single(gty.args)))

    @mfs.simple(lambda _, ctx, rty: isinstance(rty, type) and issubclass(rty, collections.abc.Iterable))
    def _build_concrete(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        return IterableMarshaler(ctx.make(ta.Any))


@dc.dataclass(frozen=True)
class IterableUnmarshaler(Unmarshaler):
    ctor: ta.Callable[[ta.Iterable[ta.Any]], ta.Iterable]
    e: Unmarshaler

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Iterable:
        return self.ctor(map(functools.partial(self.e.unmarshal, ctx), check.isinstance(v, collections.abc.Iterable)))


class IterableUnmarshalerFactory(UnmarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: isinstance(rty, rfl.Generic) and issubclass(rty.cls, collections.abc.Iterable))
    def _build_generic(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        gty = check.isinstance(rty, rfl.Generic)
        cty = DEFAULT_ITERABLE_CONCRETE_TYPES.get(gty.cls, gty.cls)  # noqa
        return IterableUnmarshaler(cty, ctx.make(check.single(gty.args)))

    @mfs.simple(lambda _, ctx, rty: isinstance(rty, type) and issubclass(rty, collections.abc.Iterable))
    def _build_concrete(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        return IterableUnmarshaler(check.isinstance(rty, type), ctx.make(ta.Any))



import abc
import contextlib
import typing as ta

from .. import check
from .. import dataclasses as dc
from .. import lang
from .bindings import Binding
from .elements import Element
from .injector import AsyncInjector
from .keys import Key
from .keys import as_key
from .providers import Provider
from .types import Scope


if ta.TYPE_CHECKING:
    from . import injector as _injector
    from . import maysync as _maysync
    from . import sync as _sync
else:
    _injector = lang.proxy_import('.injector', __package__)
    _maysync = lang.proxy_import('.maysync', __package__)
    _sync = lang.proxy_import('.sync', __package__)


##


SCOPE_ALIASES: dict[str, Scope] = {}


@dc.dataclass(frozen=True)
@dc.extra_class_params(cache_hash=True)
class ScopeBinding(Element, lang.Final):
    scope: Scope = dc.xfield(coerce=check.of_isinstance(Scope))


def bind_scope(sc: Scope) -> Element:
    return ScopeBinding(sc)


##


class Singleton(Scope, lang.Singleton, lang.Final):
    pass


SCOPE_ALIASES['singleton'] = Singleton()


##


class ThreadScope(Scope, lang.Singleton, lang.Final):
    pass


SCOPE_ALIASES['thread'] = ThreadScope()


##


@dc.dataclass(frozen=True)
@dc.extra_class_params(cache_hash=True)
class SeededScope(Scope, lang.Final):
    tag: ta.Any = dc.xfield(coerce=check.not_none)

    class Manager(lang.Abstract):
        @abc.abstractmethod
        def __call__(self, seeds: ta.Mapping[Key, ta.Any]) -> ta.AsyncContextManager[None]:
            raise NotImplementedError


@dc.dataclass(frozen=True)
@dc.extra_class_params(cache_hash=True)
class ScopeSeededProvider(Provider):
    ss: SeededScope = dc.xfield(coerce=check.of_isinstance(SeededScope))
    key: Key = dc.xfield(coerce=check.of_isinstance(Key))


def bind_scope_seed(k: ta.Any, ss: SeededScope) -> Element:
    k = as_key(k)
    return Binding(k, ScopeSeededProvider(ss, k))


##


@contextlib.asynccontextmanager
async def async_enter_seeded_scope(
        i: '_injector.AsyncInjector',
        ss: SeededScope,
        keys: ta.Mapping[Key, ta.Any],
) -> ta.AsyncGenerator[None]:
    async with (await i.provide(Key(SeededScope.Manager, tag=ss)))(keys):
        yield


def enter_seeded_scope(
        i: '_sync.Injector',
        ss: SeededScope,
        keys: ta.Mapping[Key, ta.Any],
) -> ta.ContextManager[None]:
    return lang.run_maysync_context_manager(async_enter_seeded_scope(
        i[AsyncInjector],
        ss,
        keys,
    ))


def maysync_enter_seeded_scope(
        i: '_maysync.MaysyncInjector',
        ss: SeededScope,
        keys: ta.Mapping[Key, ta.Any],
) -> ta.ContextManager[None]:
    return lang.sync_async_with(async_enter_seeded_scope(
        i[AsyncInjector],
        ss,
        keys,
    ))



"""
FIXME:
 - don't base64 by default, only for json-esque targets
"""
import base64
import typing as ta

from ... import check
from ... import dataclasses as dc
from ..base.contexts import MarshalContext
from ..base.contexts import UnmarshalContext
from ..base.types import Marshaler
from ..base.types import Unmarshaler
from ..base.values import Value
from ..factories.typemap import TypeMapMarshalerFactory
from ..factories.typemap import TypeMapUnmarshalerFactory


T = ta.TypeVar('T')


##


@dc.dataclass(frozen=True)
class Base64MarshalerUnmarshaler(Marshaler, Unmarshaler, ta.Generic[T]):
    ty: type[T]

    def marshal(self, ctx: MarshalContext, o: bytes) -> str:
        return base64.b64encode(o).decode()

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> T:
        return self.ty(base64.b64decode(check.isinstance(v, str).encode()))  # type: ignore


BASE64_TYPES = (
    bytes,
    bytearray,
)

BASE64_MARSHALER_FACTORY = TypeMapMarshalerFactory({
    ty: Base64MarshalerUnmarshaler(ty) for ty in BASE64_TYPES
})

BASE64_UNMARSHALER_FACTORY = TypeMapUnmarshalerFactory({
    ty: Base64MarshalerUnmarshaler(ty) for ty in BASE64_TYPES
})



import sys as _sys
import typing as _ta

from .. import lang as _lang  # noqa


with _lang.auto_proxy_init(globals()):
    ##

    from .asyncs import (  # noqa
        AbstractAsyncSubprocesses,
    )

    from .base import (  # noqa
        SubprocessChannelOption,

        VerboseCalledProcessError,
        BaseSubprocesses,
    )

    from .maysync import (  # noqa
        MaysyncSubprocesses,
    )

    from .run import (  # noqa
        SubprocessRunOutput,
        SubprocessRun,
        SubprocessRunnable,
    )

    from .sync import (  # noqa
        AbstractSubprocesses,

        Subprocesses,

        subprocesses,
    )

    from .utils import (  # noqa
        subprocess_close as close,
    )

    from .wrap import (  # noqa
        subprocess_shell_wrap_exec as shell_wrap_exec,
        subprocess_maybe_shell_wrap_exec as maybe_shell_wrap_exec,
    )


##


if _ta.TYPE_CHECKING:
    from . import sync as _sync  # noqa
else:
    _sync = _lang.proxy_import('.sync', __package__)


def run(
        *cmd: str,
        input: _ta.Any = None,  # noqa
        timeout: '_lang.TimeoutLike' = None,
        check: bool = False,
        capture_output: bool | None = None,
        **kwargs: _ta.Any,
) -> 'SubprocessRunOutput':
    return _sync.subprocesses.run(
        *cmd,
        input=input,
        timeout=timeout,
        check=check,
        capture_output=capture_output,
        **kwargs,
    )


def check_call(
        *cmd: str,
        stdout: _ta.Any = _sys.stderr,
        **kwargs: _ta.Any,
) -> None:
    return _sync.subprocesses.check_call(
        *cmd,
        stdout=stdout,
        **kwargs,
    )


def check_output(
        *cmd: str,
        **kwargs: _ta.Any,
) -> bytes:
    return _sync.subprocesses.check_output(
        *cmd,
        **kwargs,
    )


def check_output_str(
        *cmd: str,
        **kwargs: _ta.Any,
) -> str:
    return _sync.subprocesses.check_output_str(
        *cmd,
        **kwargs,
    )


def try_call(
        *cmd: str,
        **kwargs: _ta.Any,
) -> bool:
    return _sync.subprocesses.try_call(
        *cmd,
        **kwargs,
    )


def try_output(
        *cmd: str,
        **kwargs: _ta.Any,
) -> bytes | None:
    return _sync.subprocesses.try_output(
        *cmd,
        **kwargs,
    )


def try_output_str(
        *cmd: str,
        **kwargs: _ta.Any,
) -> str | None:
    return _sync.subprocesses.try_output_str(
        *cmd,
        **kwargs,
    )



import abc
import typing as ta

from .. import lang


T = ta.TypeVar('T')
U = ta.TypeVar('U')


##


class Fn(lang.Abstract, ta.Generic[T]):
    @abc.abstractmethod
    def __call__(self, *args: ta.Any, **kwargs: ta.Any) -> T:
        raise NotImplementedError

    def pipe(self, fn: ta.Callable[..., U], *args: ta.Any, **kwargs: ta.Any) -> 'Fn[U]':
        return pipe(self, bind(fn, *args, **kwargs))

    def __or__(self, fn: ta.Callable[..., U]) -> 'Fn[U]':
        return pipe(self, fn)

    def __ror__(self, fn: ta.Callable[..., U]) -> 'Fn[U]':
        return pipe(fn, self)

    def apply(self, fn: ta.Callable[[T], ta.Any], *args: ta.Any, **kwargs: ta.Any) -> 'Fn[T]':
        return pipe(self, apply(bind(fn, *args, **kwargs)))

    def __and__(self, fn: ta.Callable[[T], ta.Any]) -> 'Fn[T]':
        return pipe(self, apply(fn))

    def __rand__(self, fn: ta.Callable[[T], ta.Any]) -> 'Fn[T]':
        return pipe(fn, apply(self))


##


class Bind(Fn[T]):
    def __init__(self, fn: ta.Callable[..., T], *args: ta.Any, **kwargs: ta.Any) -> None:
        super().__init__()

        if Ellipsis not in args and Ellipsis not in kwargs:
            args += (Ellipsis,)
        self._fn = fn
        self._args = args
        self._kwargs = kwargs

    def __call__(self, *args: ta.Any, **kwargs: ta.Any) -> T:
        fa: list = []
        for a in self._args:
            if a is Ellipsis:
                fa.extend(args)
            else:
                fa.append(a)

        fkw = {}
        for k, v in self._kwargs.items():
            if v is Ellipsis:
                if len(args) != 1:
                    raise ValueError(args)
                fkw[k] = args[0]
            else:
                fkw[k] = v
        fkw.update(kwargs)

        return self._fn(*fa, **fkw)


bind = Bind


##


class Pipe(Fn[T]):
    def __init__(self, lfns: ta.Sequence[ta.Callable], rfn: ta.Callable[..., T]) -> None:
        super().__init__()

        self._lfn, *self._rfns = [*lfns, rfn]

    def __call__(self, *args: ta.Any, **kwargs: ta.Any) -> T:
        o = self._lfn(*args, **kwargs)
        for fn in self._rfns:
            o = fn(o)
        return o


def pipe(*fns: ta.Callable) -> Pipe:
    *lfns, rfn = fns
    return Pipe(lfns, rfn)


##


class Apply(Fn[T]):
    def __init__(self, *fns: ta.Callable[[T], ta.Any]) -> None:
        super().__init__()

        self._fns = fns

    def __call__(self, o: T) -> T:  # noqa
        for fn in self._fns:
            fn(o)
        return o


apply = Apply



import typing as ta


##


def camel_case(name: str, *, lower: bool = False) -> str:
    if not name:
        return ''
    s = ''.join(map(str.capitalize, name.split('_')))  # noqa
    if lower:
        s = s[0].lower() + s[1:]
    return s


def snake_case(name: str) -> str:
    uppers: list[int | None] = [i for i, c in enumerate(name) if c.isupper()]
    return '_'.join([name[l:r].lower() for l, r in zip([None, *uppers], [*uppers, None])]).strip('_')


##


def is_dunder(name: str) -> bool:
    return (
        name[:2] == name[-2:] == '__' and
        name[2:3] != '_' and
        name[-3:-2] != '_' and
        len(name) > 4
    )


def is_sunder(name: str) -> bool:
    return (
        name[0] == name[-1] == '_' and
        name[1:2] != '_' and
        name[-2:-1] != '_' and
        len(name) > 2
    )


##


def strip_with_newline(s: str) -> str:
    if not s:
        return ''
    return s.strip() + '\n'


@ta.overload
def split_keep_delimiter(s: str, d: str) -> str:
    ...


@ta.overload
def split_keep_delimiter(s: bytes, d: bytes) -> bytes:
    ...


def split_keep_delimiter(s, d):
    ps = []
    i = 0
    while i < len(s):
        if (n := s.find(d, i)) < i:
            ps.append(s[i:])
            break
        ps.append(s[i:n + 1])
        i = n + 1
    return ps


##


FORMAT_NUM_BYTES_SUFFIXES: ta.Sequence[str] = ['B', 'kB', 'MB', 'GB', 'TB', 'PB', 'EB']


def format_num_bytes(num_bytes: int) -> str:
    for i, suffix in enumerate(FORMAT_NUM_BYTES_SUFFIXES):
        value = num_bytes / 1024 ** i
        if num_bytes < 1024 ** (i + 1):
            if value.is_integer():
                return f'{int(value)}{suffix}'
            else:
                return f'{value:.2f}{suffix}'

    return f'{num_bytes / 1024 ** (len(FORMAT_NUM_BYTES_SUFFIXES) - 1):.2f}{FORMAT_NUM_BYTES_SUFFIXES[-1]}'



# @omlish-lite
# ruff: noqa: UP006 UP007
"""
TODO:
 - handle secrets (but they're strs..)
"""
import http.client
import typing as ta

from ..lite.cached import cached_nullary
from ..lite.cached import cached_property
from ..lite.check import check


StrOrBytes = ta.Union[str, bytes]  # ta.TypeAlias


##


CanHttpHeaders = ta.Union[  # ta.TypeAlias  # omlish-amalg-typing-no-move
    'HttpHeaders',

    http.client.HTTPMessage,

    ta.Mapping[str, str],
    ta.Mapping[str, ta.Sequence[str]],
    ta.Mapping[str, ta.Union[str, ta.Sequence[str]]],

    ta.Mapping[bytes, bytes],
    ta.Mapping[bytes, ta.Sequence[bytes]],
    ta.Mapping[bytes, ta.Union[bytes, ta.Sequence[bytes]]],

    ta.Mapping[StrOrBytes, StrOrBytes],
    ta.Mapping[StrOrBytes, ta.Sequence[StrOrBytes]],
    ta.Mapping[StrOrBytes, ta.Union[StrOrBytes, ta.Sequence[StrOrBytes]]],

    ta.Sequence[ta.Tuple[str, str]],
    ta.Sequence[ta.Tuple[bytes, bytes]],
    ta.Sequence[ta.Tuple[StrOrBytes, StrOrBytes]],
]


class DuplicateHttpHeaderError(Exception):
    pass


class HttpHeaders:
    def __init__(self, src: CanHttpHeaders) -> None:
        super().__init__()

        if isinstance(src, HttpHeaders):
            check.is_(src, self)
            return

        # TODO: optimized storage, 'use-whats-given'
        lst: ta.List[ta.Tuple[bytes, bytes]] = []
        if isinstance(src, http.client.HTTPMessage):
            lst = [(self._as_bytes(k), self._as_bytes(v)) for k, v in src.items()]

        elif isinstance(src, ta.Mapping):
            for k, v in src.items():
                if isinstance(v, (str, bytes)):
                    lst.append((self._as_bytes(k), self._as_bytes(v)))
                else:
                    for e in v:
                        lst.append((self._as_bytes(k), self._as_bytes(e)))

        elif isinstance(src, (str, bytes)):  # type: ignore
            raise TypeError(src)

        elif isinstance(src, ta.Sequence):
            for t in src:
                if isinstance(t, (str, bytes)):
                    raise TypeError(t)

                k, v = t
                lst.append((self._as_bytes(k), self._as_bytes(v)))

        else:
            raise TypeError(src)

        self._lst = lst

    def __new__(cls, obj: CanHttpHeaders) -> 'HttpHeaders':
        if isinstance(obj, HttpHeaders):
            return obj

        return super().__new__(cls)

    #

    # https://github.com/pgjones/hypercorn/commit/13f385be7277f407a9a361c958820515e16e217e
    ENCODING: ta.ClassVar[str] = 'latin1'

    @classmethod
    def _as_bytes(cls, o: StrOrBytes) -> bytes:
        if isinstance(o, bytes):
            return o
        elif isinstance(o, str):
            return o.encode(cls.ENCODING)
        else:
            raise TypeError(o)

    #

    @cached_nullary
    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({{{", ".join(repr(k) for k in self.single_str_dct)}}})'

    #

    @property
    def raw(self) -> ta.Sequence[ta.Tuple[bytes, bytes]]:
        return self._lst

    @classmethod
    def _as_key(cls, o: StrOrBytes) -> bytes:
        return cls._as_bytes(o).lower()

    @cached_property
    def normalized(self) -> ta.Sequence[ta.Tuple[bytes, bytes]]:
        return [(self._as_key(k), v) for k, v in self._lst]

    #

    @cached_property
    def multi_dct(self) -> ta.Mapping[bytes, ta.Sequence[bytes]]:
        d: ta.Dict[bytes, ta.List[bytes]] = {}
        for k, v in self.normalized:
            try:
                l = d[k]
            except KeyError:
                l = d[k] = []
            l.append(v)
        return d

    @cached_property
    def single_dct(self) -> ta.Mapping[bytes, bytes]:
        return {k: v[0] for k, v in self.multi_dct.items() if len(v) == 1}

    @cached_property
    def strict_dct(self) -> ta.Mapping[bytes, bytes]:
        d: ta.Dict[bytes, bytes] = {}
        for k, v in self.normalized:
            if k in d:
                if True:
                    raise DuplicateHttpHeaderError(k)
            else:
                d[k] = v
        return d

    #

    @cached_property
    def strs(self) -> ta.Sequence[ta.Tuple[str, str]]:
        return tuple((k.decode(self.ENCODING), v.decode(self.ENCODING)) for k, v in self.normalized)

    @cached_property
    def multi_str_dct(self) -> ta.Mapping[str, ta.Sequence[str]]:
        d: ta.Dict[str, ta.List[str]] = {}
        for k, v in self.strs:
            try:
                l = d[k]
            except KeyError:
                l = d[k] = []
            l.append(v)
        return d

    @cached_property
    def single_str_dct(self) -> ta.Mapping[str, str]:
        return {k: v[0] for k, v in self.multi_str_dct.items() if len(v) == 1}

    @cached_property
    def strict_str_dct(self) -> ta.Mapping[str, str]:
        d: ta.Dict[str, str] = {}
        for k, v in self.strs:
            if k in d:
                if True:
                    raise DuplicateHttpHeaderError(k)
            else:
                d[k] = v
        return d

    #

    def __bool__(self) -> bool:
        return bool(self._lst)

    def __len__(self) -> int:
        return len(self._lst)

    def __iter__(self) -> ta.Iterator[ta.Tuple[bytes, bytes]]:
        return iter(self._lst)

    @ta.overload
    def __getitem__(self, item: str) -> ta.Sequence[str]:
        ...

    @ta.overload
    def __getitem__(self, item: bytes) -> ta.Sequence[bytes]:
        ...

    @ta.overload
    def __getitem__(self, item: int) -> ta.Tuple[StrOrBytes, StrOrBytes]:
        ...

    @ta.overload
    def __getitem__(self, item: slice) -> ta.Sequence[ta.Tuple[StrOrBytes, StrOrBytes]]:
        ...

    def __getitem__(self, item):
        if isinstance(item, (int, slice)):
            return self._lst[item]
        elif isinstance(item, str):
            return self.multi_str_dct[item.lower()]
        elif isinstance(item, bytes):
            return self.multi_dct[self._as_key(item)]
        else:
            raise TypeError(item)

    def keys(self) -> ta.Iterable[bytes]:
        return self.multi_dct.keys()

    def items(self) -> ta.Iterable[ta.Tuple[bytes, bytes]]:
        return self._lst

    #

    def update(
            self,
            *items: ta.Tuple[bytes, bytes],  # FIXME: all arg types
            override: bool = False,
    ) -> 'HttpHeaders':
        if override:
            nks = {k.lower() for k, v in items}
            src = [(k, v) for k, v in self.items() if k.lower() not in nks]
        else:
            src = list(self.items())
        return HttpHeaders([
            *src,
            *items,
        ])


def headers(src: CanHttpHeaders) -> HttpHeaders:
    return HttpHeaders(src)



import dataclasses as dc
import typing as ta


T = ta.TypeVar('T')


##


def deep_replace(o: T, *args: str | ta.Callable[[ta.Any], ta.Mapping[str, ta.Any]]) -> T:
    if not args:
        return o
    elif len(args) == 1:
        return dc.replace(o, **args[0](o))  # type: ignore
    else:
        return dc.replace(o, **{args[0]: deep_replace(getattr(o, args[0]), *args[1:])})  # type: ignore



"""
TODO:
 - option to coordinate with objects and omit if empty / render unboxed
"""
import dataclasses as dc
import typing as ta

from ... import check
from ... import lang
from ... import reflect as rfl
from ...funcs import match as mfs
from ..base.contexts import MarshalContext
from ..base.contexts import UnmarshalContext
from ..base.types import Marshaler
from ..base.types import Unmarshaler
from ..base.values import Value
from ..factories.match import MarshalerFactoryMatchClass
from ..factories.match import UnmarshalerFactoryMatchClass


##


@dc.dataclass(frozen=True)
class MaybeMarshaler(Marshaler):
    e: Marshaler

    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        m: lang.Maybe = o
        if m.present:
            return [self.e.marshal(ctx, m.must())]
        else:
            return []


class MaybeMarshalerFactory(MarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: isinstance(rty, rfl.Generic) and rty.cls is lang.Maybe)
    def _build_generic(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        gty = check.isinstance(rty, rfl.Generic)
        return MaybeMarshaler(ctx.make(check.single(gty.args)))

    @mfs.simple(lambda _, ctx, rty: isinstance(rty, type) and issubclass(rty, lang.Maybe))
    def _build_concrete(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        return MaybeMarshaler(ctx.make(ta.Any))


@dc.dataclass(frozen=True)
class MaybeUnmarshaler(Unmarshaler):
    e: Unmarshaler

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        if v:
            return lang.just(self.e.unmarshal(ctx, check.single(v)))  # type: ignore
        else:
            return lang.empty()


class MaybeUnmarshalerFactory(UnmarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: isinstance(rty, rfl.Generic) and rty.cls is lang.Maybe)
    def _build_generic(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        gty = check.isinstance(rty, rfl.Generic)
        return MaybeUnmarshaler(ctx.make(check.single(gty.args)))

    @mfs.simple(lambda _, ctx, rty: isinstance(rty, type) and issubclass(rty, lang.Maybe))
    def _build_concrete(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        return MaybeUnmarshaler(ctx.make(ta.Any))



# @omlish-lite
import logging
import typing as ta

from .base import Logger
from .std.loggers import StdLogger


##


def get_module_logger(mod_globals: ta.Mapping[str, ta.Any]) -> Logger:
    return StdLogger(logging.getLogger(mod_globals.get('__name__')))  # noqa



import dataclasses as dc
import inspect
import json
import math
import re
import typing as ta

from .errors import ArityError
from .errors import JmespathError
from .errors import JmespathTypeError
from .errors import JmespathValueError
from .errors import UnknownFunctionError
from .errors import VariadicArityError


T = ta.TypeVar('T')


##


JmespathType: ta.TypeAlias = ta.Literal[
    'boolean',
    'array',
    'object',
    'null',
    'string',
    'number',
    'expref',
]


PyType = ta.NewType('PyType', str)


def pytype_of(o: ta.Any) -> PyType:
    return PyType(type(o).__name__)


TYPES_MAP: ta.Mapping[PyType, JmespathType] = {
    PyType('bool'): 'boolean',
    PyType('list'): 'array',
    PyType('dict'): 'object',
    PyType('NoneType'): 'null',
    PyType('unicode'): 'string',
    PyType('str'): 'string',
    PyType('float'): 'number',
    PyType('int'): 'number',
    PyType('long'): 'number',
    PyType('OrderedDict'): 'object',
    PyType('_Projection'): 'array',
    PyType('_Expression'): 'expref',
}


REVERSE_TYPES_MAP: ta.Mapping[JmespathType, ta.Sequence[PyType]] = {
    'boolean': [
        PyType('bool'),
    ],
    'array': [
        PyType('list'),
        PyType('_Projection'),
    ],
    'object': [
        PyType('dict'),
        PyType('OrderedDict'),
    ],
    'null': [
        PyType('NoneType'),
    ],
    'string': [
        PyType('unicode'),
        PyType('str'),
    ],
    'number': [
        PyType('float'),
        PyType('int'),
    ],
    'expref': [
        PyType('_Expression'),
    ],
}


ArrayParameterType: ta.TypeAlias = ta.Literal[
    'array-string',
    'array-number',
]

ParameterType: ta.TypeAlias = JmespathType | ArrayParameterType


class Parameter(ta.TypedDict):
    type: ta.NotRequired[ParameterType]
    types: ta.NotRequired[ta.Sequence[ParameterType]]
    variadic: ta.NotRequired[bool]
    optional: ta.NotRequired[bool]


Signature: ta.TypeAlias = ta.Sequence[Parameter]


@dc.dataclass(frozen=True)
class Function:
    function: ta.Callable
    signature: Signature


def signature(*params: Parameter):
    def _record_signature(func):
        func.signature = params
        return func
    return _record_signature


class Functions(ta.Protocol):
    def call_function(self, function_name, resolved_args): ...


class FunctionsClass:
    _function_table: ta.ClassVar[ta.Mapping[str, Function]] = {}  # noqa

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        cls._populate_function_table()

    @classmethod
    def _populate_function_table(cls) -> None:
        function_table: dict[str, Function] = {}

        # Any method with a @signature decorator that also starts with "_func_" is registered as a function.
        # _func_max_by -> max_by function.
        for name, method in inspect.getmembers(cls, predicate=inspect.isfunction):
            if not name.startswith('_func_'):
                continue

            sig = getattr(method, 'signature', None)
            if sig is not None:
                function_table[name[6:]] = Function(
                    method,
                    sig,
                )

        cls._function_table = function_table

    #

    def call_function(self, function_name: str, resolved_args: ta.Sequence[ta.Any]) -> ta.Any:
        try:
            spec = self._function_table[function_name]
        except KeyError:
            raise UnknownFunctionError(f'Unknown function: {function_name}()')  # noqa

        self._validate_arguments(resolved_args, spec.signature, function_name)
        return spec.function.__get__(self, self.__class__)(*resolved_args)  # noqa

    def _validate_arguments(self, args: ta.Sequence[ta.Any], sig: Signature, function_name: str) -> None:
        if len(sig) == 0:
            self._type_check(args, sig, function_name)
            return

        required_arguments_count = len([
            param for param in sig if param and (not param.get('optional') or not param['optional'])
        ])
        optional_arguments_count = len([
            param for param in sig if param and param.get('optional') and param['optional']
        ])
        has_variadic = sig[-1].get('variadic') if sig is not None else False

        if has_variadic:
            if len(args) < len(sig):
                raise VariadicArityError(len(sig), len(args), function_name)

        elif optional_arguments_count > 0:
            if (
                    len(args) < required_arguments_count or
                    len(args) > (required_arguments_count + optional_arguments_count)
            ):
                raise ArityError(len(sig), len(args), function_name)

        elif len(args) != required_arguments_count:
            raise ArityError(len(sig), len(args), function_name)

        self._type_check(args, sig, function_name)

    def _type_check(self, actual: ta.Sequence[ta.Any], sig: Signature, function_name: str) -> None:
        for i in range(min(len(sig), len(actual))):
            allowed_types = self._get_allowed_types_from_signature(sig[i])
            if allowed_types:
                self._type_check_single(actual[i], allowed_types, function_name)

    def _get_allowed_types_from_signature(self, spec: Parameter) -> ta.Sequence[ParameterType]:
        # signature supports monotype {'type': 'type-name'}## or multiple types {'types': ['type1-name', 'type2-name']}
        return [
            *([st] if (st := spec.get('type')) is not None else []),
            *spec.get('types', []),
        ]

    def _type_check_single(self, current: ta.Any, types: ta.Sequence[ParameterType], function_name: str) -> None:
        # Type checking involves checking the top level type, and in the case of arrays, potentially checking the types
        # of each element.
        allowed_types, allowed_element_types = self._get_allowed_pytypes(types)

        # We're not using isinstance() on purpose. The type model for jmespath does not map 1-1 with python types
        # (booleans are considered integers in python for example).
        pytype = pytype_of(current)
        if pytype not in allowed_types:
            raise JmespathTypeError(
                function_name,
                current,
                self._convert_to_jmespath_type(pytype),
                types,
            )

        # If we're dealing with a list type, we can have additional restrictions on the type of the list elements (for
        # example a function can require a list of numbers or a list of strings). Arrays are the only types that can
        # have element types.
        if allowed_element_types:
            self._element_type_check(
                current,
                allowed_element_types,
                types,
                function_name,
            )

    class _AllowedPytypes(ta.NamedTuple):
        types: ta.Sequence[PyType]
        element_types: ta.Sequence[ta.Sequence[PyType]]

    def _get_allowed_pytypes(self, types: ta.Iterable[ParameterType]) -> _AllowedPytypes:
        allowed_types: list = []
        allowed_element_types: list = []

        t: str
        for t in types:
            if '-' in t:
                t, et = t.split('-', 1)
                allowed_element_types.append(REVERSE_TYPES_MAP[ta.cast(JmespathType, et)])

            allowed_types.extend(REVERSE_TYPES_MAP[ta.cast(JmespathType, t)])

        return FunctionsClass._AllowedPytypes(allowed_types, allowed_element_types)

    def _element_type_check(
            self,
            current: ta.Sequence[ta.Any],
            allowed_element_types: ta.Sequence[ta.Sequence[PyType]],
            types: ta.Sequence[ParameterType],
            function_name: str,
    ) -> None:
        if len(allowed_element_types) == 1:
            # The easy case, we know up front what type we need to validate.
            ets = allowed_element_types[0]
            for element in current:
                pytype = pytype_of(element)
                if pytype not in ets:
                    raise JmespathTypeError(function_name, element, pytype, types)

        elif len(allowed_element_types) > 1 and current:
            # Dynamic type validation. Based on the first type we see, we validate that the remaining types match.
            first = pytype_of(current[0])
            for element_types in allowed_element_types:
                if first in element_types:
                    allowed = element_types
                    break
            else:
                raise JmespathTypeError(function_name, current[0], first, types)

            for element in current:
                pytype = pytype_of(element)
                if pytype not in allowed:
                    raise JmespathTypeError(function_name, element, pytype, types)

    #

    def _convert_to_jmespath_type(self, pytype: PyType) -> JmespathType | ta.Literal['unknown']:
        return TYPES_MAP.get(pytype, 'unknown')

    def _ensure_integer(
            self,
            func_name: str,
            param_name: str,
            param_value: ta.Any,
    ) -> None:
        if param_value is not None:
            if int(param_value) != param_value:
                raise JmespathValueError(
                    func_name,
                    param_value,
                    'integer',
                )

    def _ensure_non_negative_integer(
            self,
            func_name: str,
            param_name: str,
            param_value: ta.Any,
    ) -> None:
        if param_value is not None:
            if int(param_value) != param_value or int(param_value) < 0:
                raise JmespathValueError(
                    func_name,
                    param_name,
                    'non-negative integer',
                )


##


class TypeFunctions(FunctionsClass):
    @signature({'types': []})
    def _func_type(self, arg):
        if isinstance(arg, str):
            return 'string'
        elif isinstance(arg, bool):
            return 'boolean'
        elif isinstance(arg, list):
            return 'array'
        elif isinstance(arg, dict):
            return 'object'
        elif isinstance(arg, (float, int)):
            return 'number'
        elif arg is None:
            return 'null'
        else:
            return None

    @signature({'types': [], 'variadic': True})
    def _func_not_null(self, *arguments):
        for argument in arguments:
            if argument is not None:
                return argument
        return None

    @signature({'types': []})
    def _func_to_array(self, arg):
        if isinstance(arg, list):
            return arg
        else:
            return [arg]

    @signature({'types': []})
    def _func_to_string(self, arg):
        if isinstance(arg, str):
            return arg
        else:
            return json.dumps(arg, separators=(',', ':'), default=str)

    @signature({'types': []})
    def _func_to_number(self, arg):
        if isinstance(arg, (list, dict, bool)):
            return None

        elif arg is None:
            return None

        elif isinstance(arg, (int, float)):
            return arg

        else:
            try:
                return int(arg)
            except ValueError:
                try:
                    return float(arg)
                except ValueError:
                    return None


class ContainerFunctions(FunctionsClass):
    @signature({'types': ['array', 'string']}, {'types': [], 'variadic': True})
    def _func_contains(self, subject, *searches):
        return any(search in subject for search in searches)

    @signature({'types': ['array', 'string']}, {'types': [], 'variadic': True})
    def _func_in(self, subject, *searches):
        return subject in searches

    @signature({'types': ['string', 'array', 'object']})
    def _func_length(self, arg):
        return len(arg)

    @signature({'types': ['array', 'string']})
    def _func_reverse(self, arg):
        if isinstance(arg, str):
            return arg[::-1]
        else:
            return list(reversed(arg))

    @signature({'types': ['expref']}, {'types': ['array']})
    def _func_map(self, expref, arg):
        result = []
        for element in arg:
            result.append(expref.visit(expref.expression, element))
        return result

    @signature({'types': ['object'], 'variadic': True})
    def _func_merge(self, *arguments):
        merged = {}
        for arg in arguments:
            merged.update(arg)
        return merged

    @signature({'types': ['array-string', 'array-number']})
    def _func_sort(self, arg):
        return sorted(arg)

    @signature({'types': ['array'], 'variadic': True})
    def _func_zip(self, *arguments):
        return [list(t) for t in zip(*arguments)]


class NumberFunctions(FunctionsClass):
    @signature({'types': ['number']})
    def _func_abs(self, arg):
        return abs(arg)

    @signature({'types': ['array-number']})
    def _func_avg(self, arg):
        if arg:
            return sum(arg) / len(arg)
        else:
            return None

    @signature({'types': ['number']})
    def _func_ceil(self, arg):
        return math.ceil(arg)

    @signature({'types': ['number']})
    def _func_floor(self, arg):
        return math.floor(arg)

    @signature({'types': ['array-number', 'array-string']})
    def _func_max(self, arg):
        if arg:
            return max(arg)
        else:
            return None

    @signature({'types': ['array-number', 'array-string']})
    def _func_min(self, arg):
        if arg:
            return min(arg)
        else:
            return None

    @signature({'types': ['array-number']})
    def _func_sum(self, arg):
        return sum(arg)


class StringFunctions(FunctionsClass):
    @signature({'types': ['string']})
    def _func_lower(self, arg):
        return arg.lower()

    @signature({'types': ['string']})
    def _func_upper(self, arg):
        return arg.upper()

    @signature({'types': ['string']}, {'types': ['string']})
    def _func_ends_with(self, search, suffix):
        return search.endswith(suffix)

    @signature({'types': ['string']}, {'types': ['string']})
    def _func_starts_with(self, search, suffix):
        return search.startswith(suffix)

    @signature({'type': 'string'}, {'type': 'string', 'optional': True})
    def _func_trim(self, text, chars=None):
        if chars is None or len(chars) == 0:
            return text.strip()
        return text.strip(chars)

    @signature({'type': 'string'}, {'type': 'string', 'optional': True})
    def _func_trim_left(self, text, chars=None):
        if chars is None or len(chars) == 0:
            return text.lstrip()
        return text.lstrip(chars)

    @signature({'type': 'string'}, {'type': 'string', 'optional': True})
    def _func_trim_right(self, text, chars=None):
        if chars is None or len(chars) == 0:
            return text.rstrip()
        return text.rstrip(chars)

    @signature({'types': ['string']}, {'types': ['array-string']})
    def _func_join(self, separator, array):
        return separator.join(array)

    #

    def _find_impl(self, text, search, func, start, end):
        if len(search) == 0:
            return None
        if end is None:
            end = len(text)

        pos = func(text[start:end], search)
        if start < 0:
            start = start + len(text)

        # restrict resulting range to valid indices
        start = min(max(start, 0), len(text))
        return start + pos if pos != -1 else None

    @signature(
        {'type': 'string'},
        {'type': 'string'},
        {'type': 'number', 'optional': True},
        {'type': 'number', 'optional': True},
    )
    def _func_find_first(self, text, search, start=0, end=None):
        self._ensure_integer('find_first', 'start', start)
        self._ensure_integer('find_first', 'end', end)
        return self._find_impl(
            text,
            search,
            lambda t, s: t.find(s),
            start,
            end,
        )

    @signature(
        {'type': 'string'},
        {'type': 'string'},
        {'type': 'number', 'optional': True},
        {'type': 'number', 'optional': True})
    def _func_find_last(self, text, search, start=0, end=None):
        self._ensure_integer('find_last', 'start', start)
        self._ensure_integer('find_last', 'end', end)
        return self._find_impl(
            text,
            search,
            lambda t, s: t.rfind(s),
            start,
            end,
        )

    #

    def _pad_impl(self, func, padding):
        if len(padding) != 1:
            raise JmespathError(
                f'syntax-error: pad_right() expects $padding to have a single character, but received '
                f'`{padding}` instead.',
            )
        return func()

    @signature(
        {'type': 'string'},
        {'type': 'number'},
        {'type': 'string', 'optional': True},
    )
    def _func_pad_left(self, text, width, padding=' '):
        self._ensure_non_negative_integer('pad_left', 'width', width)
        return self._pad_impl(lambda: text.rjust(width, padding), padding)

    @signature(
        {'type': 'string'},
        {'type': 'number'},
        {'type': 'string', 'optional': True},
    )
    def _func_pad_right(self, text, width, padding=' '):
        self._ensure_non_negative_integer('pad_right', 'width', width)
        return self._pad_impl(lambda: text.ljust(width, padding), padding)

    #

    @signature(
        {'type': 'string'},
        {'type': 'string'},
        {'type': 'string'},
        {'type': 'number', 'optional': True},
    )
    def _func_replace(self, text, search, replacement, count=None):
        self._ensure_non_negative_integer(
            'replace',
            'count',
            count,
        )

        if count is not None:
            return text.replace(search, replacement, int(count))

        return text.replace(search, replacement)

    @signature(
        {'type': 'string'},
        {'type': 'string'},
        {'type': 'number', 'optional': True},
    )
    def _func_split(self, text, search, count=None):
        self._ensure_non_negative_integer(
            'split',
            'count',
            count,
        )

        if len(search) == 0:
            chars = list(text)
            if count is None:
                return chars

            head = list(chars[:count])
            tail = [''.join(chars[count:])]
            return head + tail

        if count is not None:
            return text.split(search, count)

        return text.split(search)

    #

    @signature({'types': ['string']}, {'types': ['string']})
    def _func_match(self, string, pattern):
        return re.match(pattern, string) is not None


class ObjectFunctions(FunctionsClass):
    @signature({'types': ['object']})
    def _func_items(self, arg):
        return [list(t) for t in arg.items()]

    @signature({'types': ['array']})
    def _func_from_items(self, items):
        return dict(items)

    @signature({'types': ['object']})
    def _func_keys(self, arg):
        # To be consistent with .values() should we also return the indices of a list?
        return list(arg.keys())

    @signature({'types': ['object']})
    def _func_values(self, arg):
        return list(arg.values())

    @signature({'types': ['expref']}, {'types': ['object']})
    def _func_filter_keys(self, expref, arg):
        return {k: v for k, v in arg.items() if expref.visit(expref.expression, k)}

    @signature({'types': ['expref']}, {'types': ['object']})
    def _func_filter_values(self, expref, arg):
        return {k: v for k, v in arg.items() if expref.visit(expref.expression, v)}


class KeyedFunctions(FunctionsClass):
    def _create_key_func(self, expref, allowed_types, function_name):
        def keyfunc(x):
            result = expref.visit(expref.expression, x)
            jmespath_type = self._convert_to_jmespath_type(pytype_of(result))
            # allowed_types is in terms of jmespath types, not python types.
            if jmespath_type not in allowed_types:
                raise JmespathTypeError(
                    function_name, result, jmespath_type, allowed_types)

            return result

        return keyfunc

    @signature({'types': ['array']}, {'types': ['expref']})
    def _func_sort_by(self, array, expref):
        if not array:
            return array

        # sort_by allows for the expref to be either a number of a string, so we have some special logic to handle this.
        # We evaluate the first array element and verify that it's either a string of a number.  We then create a key
        # function that validates that type, which requires that remaining array elements resolve to the same type as
        # the first element.
        required_type = self._convert_to_jmespath_type(pytype_of(expref.visit(expref.expression, array[0])))
        if required_type not in ['number', 'string']:
            raise JmespathTypeError(
                'sort_by',
                array[0],
                required_type,
                ['string', 'number'],
            )

        keyfunc = self._create_key_func(expref, [required_type], 'sort_by')
        return sorted(array, key=keyfunc)

    @signature({'types': ['array']}, {'types': ['expref']})
    def _func_min_by(self, array, expref):
        keyfunc = self._create_key_func(
            expref,
            ['number', 'string'],
            'min_by',
        )

        if array:
            return min(array, key=keyfunc)
        else:
            return None

    @signature({'types': ['array']}, {'types': ['expref']})
    def _func_max_by(self, array, expref):
        keyfunc = self._create_key_func(
            expref,
            ['number', 'string'],
            'max_by',
        )

        if array:
            return max(array, key=keyfunc)
        else:
            return None

    @signature({'types': ['array']}, {'types': ['expref']})
    def _func_group_by(self, array, expref):
        keyfunc = self._create_key_func(expref, ['null', 'string'], 'group_by')
        if not array:
            return None

        result = {}
        keys = list(dict.fromkeys([keyfunc(item) for item in array if keyfunc(item) is not None]))
        for key in keys:
            items = [item for item in array if keyfunc(item) == key]
            result.update({key: items})

        return result


class DefaultFunctions(
    KeyedFunctions,
    ObjectFunctions,
    StringFunctions,
    NumberFunctions,
    ContainerFunctions,
    TypeFunctions,
    FunctionsClass,
):
    pass



import typing as ta

from .. import lang
from .elements import Elemental
from .elements import as_elements
from .sync import Injector


if ta.TYPE_CHECKING:
    from .impl import maysync as _maysync
else:
    _maysync = lang.proxy_import('.impl.maysync', __package__)


T = ta.TypeVar('T')


##


class MaysyncInjector(Injector, lang.Abstract):
    pass


##


def create_maysync_injector(*args: Elemental) -> MaysyncInjector:
    return _maysync.create_maysync_injector(as_elements(*args))



from .building import (  # noqa
    JsonValueBuilder,
)

from .errors import (  # noqa
    JsonStreamError,
)


from .lexing import (  # noqa
    IdentTokenKind,
    ValueTokenKind,
    VALUE_TOKEN_KINDS,
    ControlTokenKind,
    SpaceTokenKind,
    TokenKind,

    ScalarValue,
    SCALAR_VALUE_TYPES,

    Position,

    Token,

    CONTROL_TOKENS,
    CONST_IDENT_VALUES,

    JsonStreamLexError,
    JsonStreamLexer,
)

from .parsing import (  # noqa
    BeginObject,
    Key,
    EndObject,
    BeginArray,
    EndArray,

    Event,
    Events,

    yield_parser_events,

    JsonStreamParseError,
    JsonStreamObject,
    JsonStreamParser,
)

from .rendering import (  # noqa
    StreamJsonRenderer,
)

from .utils import (  # noqa
    JsonStreamValueParser,

    stream_parse_values,
    stream_parse_one_value,
    stream_parse_exactly_one_value,
)



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import os.path
import typing as ta


##


@ta.overload
def path_dirname(p: str, n: int = 1) -> str:
    ...


@ta.overload
def path_dirname(p: bytes, n: int = 1) -> bytes:
    ...


def path_dirname(p, n=1):
    if isinstance(p, bytes):
        sep: ta.Any = b'/'
    else:
        sep = '/'
    p = os.fspath(p)
    i = -1
    for _ in range(n):
        i = p.rindex(sep, 0, i)
    head = p[:i + 1]
    if head and head != sep * len(head):
        head = head.rstrip(sep)
    return head


def abs_real_path(p: str) -> str:
    return os.path.abspath(os.path.realpath(p))


def is_path_in_dir(base_dir: str, target_path: str) -> bool:
    base_dir = abs_real_path(base_dir)
    target_path = abs_real_path(target_path)

    return target_path.startswith(base_dir + os.path.sep)


def relative_symlink(
        src: str,
        dst: str,
        *,
        target_is_directory: bool = False,
        dir_fd: ta.Optional[int] = None,
        make_dirs: bool = False,
        **kwargs: ta.Any,
) -> None:
    if make_dirs:
        os.makedirs(os.path.dirname(dst), exist_ok=True)

    os.symlink(
        os.path.relpath(src, os.path.dirname(dst)),
        dst,
        target_is_directory=target_is_directory,
        dir_fd=dir_fd,
        **kwargs,
    )



"""
TODO:
 - quantiles via math.histogram
"""
import atexit
import contextlib
import functools
import sys
import threading
import time
import typing as ta

from .. import check
from .. import lang


T = ta.TypeVar('T')


##


class _GlobalTimer:
    def __init__(
            self,
            registry: '_GlobalTimerRegistry',
            name: str,
            *,
            clock: ta.Callable[[], float] | None = None,
            report_at_exit: bool = False,
            report_out: ta.Any | None = None,
    ) -> None:
        super().__init__()

        self._registry = registry
        self._name = name

        if clock is None:
            clock = time.monotonic
        self._clock = clock
        self._report_out = report_out

        self._lock = threading.Lock()

        self._c = 0
        self._t = 0.

        if report_at_exit:
            atexit.register(self.report)

    def report(self) -> None:
        msg = (
            f'{self._registry.module_name}::{self._name}: '
            f'{self._c} calls, '
            f'{self._t:.3f}s total'
        )

        print(
            msg,
            file=self._report_out if self._report_out is not None else sys.stderr,
        )

    @contextlib.contextmanager
    def __call__(self) -> ta.Iterator[None]:
        start = self._clock()

        try:
            yield

        finally:
            end = self._clock()
            took = end - start

            with self._lock:
                self._c += 1
                self._t += took


#


class _GlobalTimerRegistry:
    def __init__(
            self,
            module_name: str,
    ) -> None:
        super().__init__()

        self._module_name = module_name

        self._lock = threading.Lock()

        self._timers: dict[str, _GlobalTimer] = {}

    @property
    def module_name(self) -> str:
        return self._module_name

    def get_timer(
            self,
            name: str,
            **kwargs: ta.Any,
    ) -> _GlobalTimer:
        return lang.double_check_setdefault(
            self._lock,
            self._timers,
            name,
            lambda: _GlobalTimer(
                self,
                name,
                **kwargs,
            ),
        )


_GLOBAL_LOCK = threading.Lock()
_GLOBAL_ATTR = '__global_timers_registry__'


def _get_global_registry(
        globals: ta.MutableMapping[str, ta.Any],  # noqa
) -> _GlobalTimerRegistry:
    reg = lang.double_check_setdefault(
        _GLOBAL_LOCK,
        globals,
        _GLOBAL_ATTR,
        lambda: _GlobalTimerRegistry(
            globals['__name__'],
        ),
    )

    return check.isinstance(reg, _GlobalTimerRegistry)


#


@contextlib.contextmanager
def global_timer_context(
        globals: ta.MutableMapping[str, ta.Any],  # noqa
        name: str,
        **kwargs: ta.Any,
) -> ta.Iterator[None]:
    reg = _get_global_registry(globals)
    timer = reg.get_timer(name, **kwargs)
    with timer():
        yield


#


class _GlobalTimerContextWrapped:
    def __init__(
            self,
            fn: ta.Any,
            timer: _GlobalTimer,
    ) -> None:
        super().__init__()

        self._fn = fn
        self._timer = timer

        functools.update_wrapper(self, fn)

    def __get__(self, instance, owner=None):
        return self.__class__(
            self._fn.__get__(instance, owner),
            self._timer,
        )

    def __call__(self, *args, **kwargs):
        with self._timer():
            return self._fn(*args, **kwargs)


def global_timer_wrap(
        globals: ta.MutableMapping[str, ta.Any],  # noqa
        name: str,
        **kwargs: ta.Any,
) -> ta.Callable[[T], T]:
    reg = _get_global_registry(globals)
    timer = reg.get_timer(name, **kwargs)

    def inner(fn):
        return _GlobalTimerContextWrapped(fn, timer)

    return inner



import collections.abc
import dataclasses as dc
import typing as ta

from ... import check
from ... import reflect as rfl
from ...funcs import match as mfs
from ..base.contexts import MarshalContext
from ..base.contexts import UnmarshalContext
from ..base.types import Marshaler
from ..base.types import Unmarshaler
from ..base.values import Value
from ..factories.match import MarshalerFactoryMatchClass
from ..factories.match import UnmarshalerFactoryMatchClass


##


DEFAULT_MAPPING_CONCRETE_TYPES: dict[type[collections.abc.Mapping], type[collections.abc.Mapping]] = {
    collections.abc.Mapping: dict,  # type: ignore
    collections.abc.MutableMapping: dict,  # type: ignore
}


@dc.dataclass(frozen=True)
class MappingMarshaler(Marshaler):
    ke: Marshaler
    ve: Marshaler

    def marshal(self, ctx: MarshalContext, o: ta.Mapping) -> Value:
        return {
            self.ke.marshal(ctx, uk): self.ve.marshal(ctx, uv)
            for uk, uv in check.isinstance(o, collections.abc.Mapping).items()
        }


class MappingMarshalerFactory(MarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: isinstance(rty, rfl.Generic) and issubclass(rty.cls, collections.abc.Mapping))
    def _build_generic(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        gty = check.isinstance(rty, rfl.Generic)
        kt, vt = gty.args
        return MappingMarshaler(ctx.make(kt), ctx.make(vt))

    @mfs.simple(lambda _, ctx, rty: isinstance(rty, type) and issubclass(rty, collections.abc.Mapping))
    def _build_concrete(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        return MappingMarshaler(a := ctx.make(ta.Any), a)


@dc.dataclass(frozen=True)
class MappingUnmarshaler(Unmarshaler):
    ctor: ta.Callable[[ta.Mapping[ta.Any, ta.Any]], ta.Mapping]
    ke: Unmarshaler
    ve: Unmarshaler

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Mapping:
        dct: dict = {}
        for mk, mv in check.isinstance(v, collections.abc.Mapping).items():
            dct[self.ke.unmarshal(ctx, mk)] = self.ve.unmarshal(ctx, mv)
        return self.ctor(dct)


class MappingUnmarshalerFactory(UnmarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: isinstance(rty, rfl.Generic) and issubclass(rty.cls, collections.abc.Mapping))
    def _build_generic(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        gty = check.isinstance(rty, rfl.Generic)
        cty = DEFAULT_MAPPING_CONCRETE_TYPES.get(gty.cls, gty.cls)  # noqa
        kt, vt = gty.args
        return MappingUnmarshaler(cty, ctx.make(kt), ctx.make(vt))

    @mfs.simple(lambda _, ctx, rty: isinstance(rty, type) and issubclass(rty, collections.abc.Mapping))
    def _build_concrete(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        return MappingUnmarshaler(check.isinstance(rty, type), a := ctx.make(ta.Any), a)



import types
import typing as ta

from .. import dataclasses as dc
from .. import lang
from .base import Lifecycle
from .controller import LifecycleController
from .states import LifecycleState
from .states import LifecycleStates


LifecycleT = ta.TypeVar('LifecycleT', bound='Lifecycle')
ContextManagerT = ta.TypeVar('ContextManagerT', bound=ta.ContextManager)


##


@dc.dataclass(frozen=True)
class ContextManagerLifecycle(Lifecycle, lang.Final, ta.Generic[ContextManagerT]):
    cm: ContextManagerT

    @ta.override
    def lifecycle_start(self) -> None:
        self.cm.__enter__()

    @ta.override
    def lifecycle_stop(self) -> None:
        self.cm.__exit__(None, None, None)


class LifecycleContextManager(ta.Generic[LifecycleT]):
    def __init__(self, lifecycle: LifecycleT) -> None:
        super().__init__()

        self._lifecycle = lifecycle
        self._controller = lifecycle if isinstance(lifecycle, LifecycleController) else LifecycleController(lifecycle)

    __repr__ = lang.attr_ops(lambda o: (o.lifecycle, o.state)).repr

    @property
    def lifecycle(self) -> LifecycleT:
        return self._lifecycle

    @property
    def controller(self) -> LifecycleController:
        return self._controller

    @property
    def state(self) -> LifecycleState:
        return self._controller.state

    def __enter__(self) -> ta.Self:
        try:
            self._controller.lifecycle_construct()
            self._controller.lifecycle_start()
        except Exception:
            self._controller.lifecycle_destroy()
            raise
        return self

    def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_val: BaseException | None,
            exc_tb: types.TracebackType | None,
    ) -> bool | None:
        try:
            if self._controller.state is LifecycleStates.STARTED:
                self._controller.lifecycle_stop()
        except Exception:
            self._controller.lifecycle_destroy()
            raise
        else:
            self._controller.lifecycle_destroy()
        return None



# ruff: noqa: I001
from .. import lang as _lang


with _lang.auto_proxy_init(globals()):
    ##

    from . import api  # noqa

    from .dbs import (  # noqa
        DbLoc,
        DbSpec,
        DbType,
        DbTypes,
        HostDbLoc,
        UrlDbLoc,
    )

    from .qualifiedname import (  # noqa
        QualifiedName,
        qn,
    )

    from . import queries  # noqa



"""
TODO:
 - col.TypeMap?
  - at least get_any
"""
import dataclasses as dc
import threading
import typing as ta

from ... import collections as col
from ... import lang


##


class RegistryItem(lang.Abstract):
    pass


RegistryItemT = ta.TypeVar('RegistryItemT', bound=RegistryItem)
RegistryItemU = ta.TypeVar('RegistryItemU', bound=RegistryItem)


@dc.dataclass(frozen=True)
class _KeyRegistryItems(ta.Generic[RegistryItemT]):
    key: ta.Any
    items: list[RegistryItemT] = dc.field(default_factory=list)
    item_lists_by_ty: dict[type[RegistryItemT], list[RegistryItemT]] = dc.field(default_factory=dict)

    def add(self, *items: RegistryItemT) -> None:
        for i in items:
            self.items.append(i)
            self.item_lists_by_ty.setdefault(type(i), []).append(i)


class RegistrySealedError(Exception):
    pass


class Registry(ta.Generic[RegistryItemT]):
    def __init__(
            self,
            *,
            lock: ta.Optional[threading.RLock] = None,  # noqa
    ) -> None:
        super().__init__()

        if lock is None:
            lock = threading.RLock()
        self._lock = lock

        self._dct: dict[ta.Any, _KeyRegistryItems[RegistryItemT]] = {}
        self._id_dct: ta.MutableMapping[ta.Any, _KeyRegistryItems[RegistryItemT]] = col.IdentityKeyDict()

        self._version = 0
        self._sealed = False

    #

    def is_sealed(self) -> bool:
        if self._sealed:
            return True
        with self._lock:
            return self._sealed

    def seal(self) -> ta.Self:
        if self._sealed:
            raise RegistrySealedError(self)
        with self._lock:
            self._seal()
        return self

    def _seal(self) -> None:
        if self._sealed:
            raise RegistrySealedError(self)

        self._sealed = True

    #

    def register(
            self,
            key: ta.Any,
            *items: RegistryItemT,
            identity: bool = False,
    ) -> ta.Self:
        if not items:
            return self

        with self._lock:
            if self._sealed:
                raise RegistrySealedError(self)

            dct: ta.Any = self._id_dct if identity else self._dct
            if (sr := dct.get(key)) is None:
                sr = dct[key] = _KeyRegistryItems(key)
            sr.add(*items)

            self._version += 1

        return self

    #

    def get(
            self,
            key: ta.Any,
            *,
            identity: bool | None = None,
    ) -> ta.Sequence[RegistryItem]:
        if identity is None:
            return (
                *self.get(key, identity=True),
                *self.get(key, identity=False),
            )

        dct: ta.Any = self._id_dct if identity else self._dct
        try:
            return dct[key].items
        except KeyError:
            return ()

    def get_of(
            self,
            key: ta.Any,
            item_ty: type[RegistryItemU],
            *,
            identity: bool | None = None,
    ) -> ta.Sequence[RegistryItemU]:
        if identity is None:
            return (
                *self.get_of(key, item_ty, identity=True),
                *self.get_of(key, item_ty, identity=False),
            )

        dct: ta.Any = self._id_dct if identity else self._dct
        try:
            sr = dct[key]
        except KeyError:
            return ()
        return sr.item_lists_by_ty.get(item_ty, ())



from .. import lang as _lang


with _lang.auto_proxy_init(
        globals(),
        update_exports=True,
        # eager=True,
):
    ##

    from .inspect import (  # noqa
        get_annotations,
        get_filtered_type_hints,
        has_annotations,
    )

    from .ops import (  # noqa
        get_concrete_type,
        get_underlying,
        strip_annotations,
        strip_objs,
        to_annotation,
        types_equivalent,
    )

    from .subst import (  # noqa
        ALIAS_UPDATING_GENERIC_SUBSTITUTION,
        DEFAULT_GENERIC_SUBSTITUTION,
        GenericSubstitution,
        generic_mro,
        get_generic_bases,
        get_type_var_replacements,
        replace_type_vars,
    )

    from .types import (  # noqa
        ANY,
        Annotated,
        Any,
        DEFAULT_REFLECTOR,
        Generic,
        GenericLike,
        Literal,
        NewType,
        Protocol,
        ReflectTypeError,
        Reflector,
        TYPES,
        Type,
        TypeInfo,
        Union,
        get_newtype_supertype,
        get_orig_bases,
        get_orig_class,
        get_params,
        get_type_var_bound,
        is_type,
        is_union_type,
        type_,
    )



import typing as ta

from ..headers import CanHttpHeaders
from .base import HttpRequest
from .base import HttpResponse
from .sync import HttpClient
from .urllib import UrllibHttpClient


##


def _default_client() -> HttpClient:
    return UrllibHttpClient()


def client() -> HttpClient:
    return _default_client()


def request(
        url: str,
        method: str | None = None,
        *,
        headers: CanHttpHeaders | None = None,
        data: bytes | str | None = None,

        timeout_s: float | None = None,

        check: bool = False,

        client: HttpClient | None = None,  # noqa

        **kwargs: ta.Any,
) -> HttpResponse:
    req = HttpRequest(
        url,
        method=method,

        headers=headers,
        data=data,

        timeout_s=timeout_s,

        **kwargs,
    )

    def do(cli: HttpClient) -> HttpResponse:
        return cli.request(
            req,

            check=check,
        )

    if client is not None:
        return do(client)

    else:
        with _default_client() as cli:
            return do(cli)



import dataclasses as dc
import typing as ta

from ... import lang
from .columns import Column
from .columns import Columns
from .errors import MismatchedColumnCountError


T = ta.TypeVar('T')


##


@dc.dataclass(frozen=True)
class Row(lang.Final, ta.Generic[T]):
    columns: Columns
    values: ta.Sequence[T]

    def __post_init__(self) -> None:
        if len(self.columns) != len(self.values):
            raise MismatchedColumnCountError(self.columns, self.values)

    #

    def __iter__(self) -> ta.Iterator[tuple[Column, T]]:
        return iter(zip(self.columns, self.values))

    def __len__(self) -> int:
        return len(self.values)

    def __contains__(self, item: str | int) -> bool:
        raise TypeError('Row.__contains__ is ambiguous - use .columns.__contains__ or .values.__contains__')

    def __getitem__(self, item: str | int) -> T:
        if isinstance(item, str):
            return self.values[self.columns.index(item)]
        elif isinstance(item, int):
            return self.values[item]
        else:
            raise TypeError(item)

    def get(self, name: str) -> T | None:
        if (idx := self.columns.get_index(name)) is not None:
            return self.values[idx]
        else:
            return None

    def to_dict(self) -> dict[str, ta.Any]:
        return {c.name: v for c, v in self}



import typing as ta

from .. import dataclasses as dc
from .. import lang


R = ta.TypeVar('R')

AnyLifecycleT = ta.TypeVar('AnyLifecycleT', bound='AnyLifecycle')
AnyLifecycleCallback: ta.TypeAlias = ta.Callable[[AnyLifecycleT], R]

LifecycleT = ta.TypeVar('LifecycleT', bound='Lifecycle')
LifecycleCallback: ta.TypeAlias = ta.Callable[[LifecycleT], R]

AsyncLifecycleT = ta.TypeVar('AsyncLifecycleT', bound='AsyncLifecycle')
AsyncLifecycleCallback: ta.TypeAlias = ta.Callable[[AsyncLifecycleT], R]


##


class AnyLifecycle(lang.Abstract, ta.Generic[R]):
    def lifecycle_construct(self) -> R | None:
        return None

    def lifecycle_start(self) -> R | None:
        return None

    def lifecycle_stop(self) -> R | None:
        return None

    def lifecycle_destroy(self) -> R | None:
        return None


@dc.dataclass(frozen=True, kw_only=True)
class AnyCallbackLifecycle(
    AnyLifecycle[R],
    lang.Abstract,
    ta.Generic[AnyLifecycleT, R],
):
    on_construct: AnyLifecycleCallback['AnyCallbackLifecycle[AnyLifecycleT, R]', R] | None = None
    on_start: AnyLifecycleCallback['AnyCallbackLifecycle[AnyLifecycleT, R]', R] | None = None
    on_stop: AnyLifecycleCallback['AnyCallbackLifecycle[AnyLifecycleT, R]', R] | None = None
    on_destroy: AnyLifecycleCallback['AnyCallbackLifecycle[AnyLifecycleT, R]', R] | None = None

    @ta.override
    def lifecycle_construct(self) -> R | None:
        if self.on_construct is not None:
            return self.on_construct(self)
        else:
            return None

    @ta.override
    def lifecycle_start(self) -> R | None:
        if self.on_start is not None:
            return self.on_start(self)
        else:
            return None

    @ta.override
    def lifecycle_stop(self) -> R | None:
        if self.on_stop is not None:
            return self.on_stop(self)
        else:
            return None

    @ta.override
    def lifecycle_destroy(self) -> R | None:
        if self.on_destroy is not None:
            return self.on_destroy(self)
        else:
            return None


##


class Lifecycle(AnyLifecycle[None]):
    pass


class CallbackLifecycle(
    AnyCallbackLifecycle[LifecycleT, None],
    lang.Final,
    ta.Generic[LifecycleT],
):
    pass


##


class AsyncLifecycle(AnyLifecycle[ta.Awaitable[None]]):
    pass


class CallbackAsyncLifecycle(
    AnyCallbackLifecycle[LifecycleT, ta.Awaitable[None]],
    lang.Final,
    ta.Generic[LifecycleT],
):
    pass



import types
import typing as ta

from .descriptors import unwrap_func


##


class _EmptyClass:
    pass


_OVERRIDE_NOT_NEEDED_ATTRS: ta.AbstractSet[str] = {
    *_EmptyClass.__dict__,
}


def needs_override(attr: str, obj: ta.Any) -> bool:
    if attr in _OVERRIDE_NOT_NEEDED_ATTRS:
        return False

    return True


##


def is_override(obj: ta.Any) -> bool:
    if isinstance(obj, types.FunctionType):
        fn = unwrap_func(obj)
        if getattr(fn, '__override__', False):
            return True

    return False


##


class RequiresOverrideError(TypeError):
    def __init__(
            self,
            att: str | None = None,
            cls: type | None = None,
            mro_cls: type | None = None,
            owner_cls: type | None = None,
    ) -> None:
        super().__init__(' '.join([
            'Attribute',
            *([f'{att!r}'] if att is not None else []),
            *([f'on class {cls.__qualname__}'] if cls is not None else []),
            *([f'from mro class {mro_cls.__qualname__}'] if mro_cls is not None and mro_cls is not cls else []),
            f'is not marked as a @typing.override',
            *([f'from owning class {owner_cls.__qualname__}'] if owner_cls is not None else []),
        ]))

        self.att = att
        self.cls = cls
        self.mro_cls = mro_cls
        self.owner_cls = owner_cls


class RequiresOverride:
    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        req_ovr_dct: dict[str, type] = {}
        for m_cls in reversed(cls.__mro__):
            for a, o in m_cls.__dict__.items():
                if a in req_ovr_dct and not is_override(o):
                    raise RequiresOverrideError(
                        a,
                        cls,
                        m_cls,
                        req_ovr_dct[a],
                    )

            if RequiresOverride in m_cls.__bases__:
                for a, o in m_cls.__dict__.items():
                    if needs_override(a, o):
                        req_ovr_dct.setdefault(a, m_cls)



import contextlib
import keyword
import sys
import types
import typing as ta

from .decorator import dataclass


##


_IS_PY_3_14 = sys.version_info >= (3, 14)

if _IS_PY_3_14:
    import annotationlib  # noqa

    _ANY_MARKER = object()


def make_dataclass(  # noqa
        cls_name,
        fields,
        *,
        bases=(),
        namespace=None,

        init=True,
        repr=True,  # noqa
        eq=True,
        order=False,
        unsafe_hash=False,
        frozen=False,

        match_args=True,
        kw_only=False,
        slots=False,
        weakref_slot=False,

        module=None,

        decorator=dataclass,

        #

        metadata: ta.Sequence[ta.Any] | None = None,

        reorder: bool | None = None,
        cache_hash: bool | None = None,
        generic_init: bool | None = None,
        override: bool | None = None,
        allow_dynamic_dunder_attrs: bool | None = None,

        repr_id: bool | None = None,
        terse_repr: bool | None = None,

        allow_redundant_decorator: bool | None = None,

        _frame_offset: int = 1,
):
    if decorator is not dataclass:
        raise TypeError(
            f'The `decorator` kwarg, as added in https://github.com/python/cpython/pull/122723, is not supported. '
            f'{decorator=}',
        )

    if namespace is None:
        namespace = {}

    seen = set()
    annotations = {}
    defaults = {}
    for item in fields:
        if isinstance(item, str):
            name = item
            if _IS_PY_3_14:
                tp = _ANY_MARKER
            else:
                tp = 'typing.Any'
        elif len(item) == 2:
            name, tp, = item
        elif len(item) == 3:
            name, tp, spec = item
            defaults[name] = spec
        else:
            raise TypeError(f'Invalid field: {item!r}')
        if not isinstance(name, str) or not name.isidentifier():
            raise TypeError(f'Field names must be valid identifiers: {name!r}')
        if keyword.iskeyword(name):
            raise TypeError(f'Field names must not be keywords: {name!r}')
        if name in seen:
            raise TypeError(f'Field name duplicated: {name!r}')

        seen.add(name)
        annotations[name] = tp

    if _IS_PY_3_14:
        # We initially block the VALUE format, because inside dataclass() we'll call get_annotations(), which will try
        # the VALUE format first. If we don't block, that means we'd always end up eagerly importing typing here, which
        # is what we're trying to avoid.
        value_blocked = True

        def annotate_method(format):  # noqa
            def get_any():
                match format:
                    case annotationlib.Format.STRING:
                        return 'typing.Any'
                    case annotationlib.Format.FORWARDREF:
                        typing = sys.modules.get('typing')
                        if typing is None:
                            return annotationlib.ForwardRef('Any', module='typing')
                        else:
                            return typing.Any
                    case annotationlib.Format.VALUE:
                        if value_blocked:
                            raise NotImplementedError
                        from typing import Any
                        return Any
                    case _:
                        raise NotImplementedError

            annos = {
                ann: get_any() if t is _ANY_MARKER else t
                for ann, t in annotations.items()
            }
            if format == annotationlib.Format.STRING:
                return annotationlib.annotations_to_string(annos)
            else:
                return annos

    def exec_body_callback(ns):
        ns.update(namespace)
        ns.update(defaults)
        if not _IS_PY_3_14:
            ns['__annotations__'] = annotations

    cls = types.new_class(cls_name, bases, {}, exec_body_callback)

    if _IS_PY_3_14:
        # For now, set annotations including the _ANY_MARKER.
        cls.__annotate__ = annotate_method  # type: ignore  # noqa

    if module is None:
        try:
            module = sys._getframemodulename(_frame_offset) or '__main__'  # noqa
        except AttributeError:
            with contextlib.suppress(AttributeError, ValueError):
                module = sys._getframe(_frame_offset).f_globals.get('__name__', '__main__')  # noqa
    if module is not None:
        cls.__module__ = module

    cls = decorator(
        cls,
        init=init,
        repr=repr,
        eq=eq,
        order=order,
        unsafe_hash=unsafe_hash,
        frozen=frozen,
        match_args=match_args,
        kw_only=kw_only,
        slots=slots,
        weakref_slot=weakref_slot,

        #

        metadata=metadata,

        reorder=reorder,
        cache_hash=cache_hash,
        generic_init=generic_init,
        override=override,
        allow_dynamic_dunder_attrs=allow_dynamic_dunder_attrs,

        repr_id=repr_id,
        terse_repr=terse_repr,

        allow_redundant_decorator=allow_redundant_decorator,
    )

    if _IS_PY_3_14:
        # Now that the class is ready, allow the VALUE format.
        value_blocked = False

    return cls



"""
TODO:
 - use importlib.util.resolve_name
"""
import importlib.util
import sys
import types
import typing as ta


##


def can_import(name: str, package: str | None = None) -> bool:
    try:
        spec = importlib.util.find_spec(name, package)
    except ImportError:
        return False
    else:
        return spec is not None


##


def try_import(spec: str) -> types.ModuleType | None:
    s = spec.lstrip('.')
    l = len(spec) - len(s)
    try:
        return __import__(s, globals(), level=l)
    except ImportError:
        return None


##


def resolve_import_name(name: str, package: str | None = None) -> str:
    # FIXME: importlib.util.resolve_name
    level = 0

    if name.startswith('.'):
        if not package:
            raise TypeError("the 'package' argument is required to perform a relative import for {name!r}")
        for character in name:
            if character != '.':
                break
            level += 1

    name = name[level:]

    if not isinstance(name, str):
        raise TypeError(f'module name must be str, not {type(name)}')
    if level < 0:
        raise ValueError('level must be >= 0')
    if level > 0:
        if not isinstance(package, str):
            raise TypeError('__package__ not set to a string')
        elif not package:
            raise ImportError('attempted relative import with no known parent package')
    if not name and level == 0:
        raise ValueError('Empty module name')

    if level > 0:
        bits = package.rsplit('.', level - 1)  # type: ignore
        if len(bits) < level:
            raise ImportError('attempted relative import beyond top-level package')
        base = bits[0]
        name = f'{base}.{name}' if name else base

    return name


##


def get_real_module_name(globals: ta.Mapping[str, ta.Any]) -> str:  # noqa
    module = sys.modules[globals['__name__']]

    if module.__spec__ and module.__spec__.name:
        return module.__spec__.name

    if module.__package__:
        return module.__package__

    raise RuntimeError("Can't determine real module name")



# ruff: noqa: UP045
# @omlish-lite
import abc
import contextlib
import dataclasses as dc
import typing as ta

from ...lite.abstract import Abstract
from ...lite.dataclasses import dataclass_maybe_post_init
from ...lite.dataclasses import dataclass_shallow_asdict
from .base import BaseHttpResponse
from .base import BaseHttpResponseT
from .base import HttpRequest
from .base import HttpResponse
from .base import HttpStatusError


StreamHttpResponseT = ta.TypeVar('StreamHttpResponseT', bound='StreamHttpResponse')
HttpClientT = ta.TypeVar('HttpClientT', bound='HttpClient')


##


@ta.final
@dc.dataclass(frozen=True)  # kw_only=True
class StreamHttpResponse(BaseHttpResponse):
    class Stream(ta.Protocol):
        def read(self, /, n: int = -1) -> bytes: ...

    @ta.final
    class _NullStream:
        def read(self, /, n: int = -1) -> bytes:
            raise TypeError

    stream: Stream = _NullStream()

    _closer: ta.Optional[ta.Callable[[], None]] = None

    def __post_init__(self) -> None:
        dataclass_maybe_post_init(super())
        if isinstance(self.stream, StreamHttpResponse._NullStream):
            raise TypeError(self.stream)

    def __enter__(self: StreamHttpResponseT) -> StreamHttpResponseT:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self) -> None:
        if (c := self._closer) is not None:
            c()


#


def close_response(resp: BaseHttpResponse) -> None:
    if isinstance(resp, HttpResponse):
        pass

    elif isinstance(resp, StreamHttpResponse):
        resp.close()

    else:
        raise TypeError(resp)


@contextlib.contextmanager
def closing_response(resp: BaseHttpResponseT) -> ta.Iterator[BaseHttpResponseT]:
    if isinstance(resp, HttpResponse):
        yield resp
        return

    elif isinstance(resp, StreamHttpResponse):
        with contextlib.closing(resp):
            yield resp

    else:
        raise TypeError(resp)


def read_response(resp: BaseHttpResponse) -> HttpResponse:
    if isinstance(resp, HttpResponse):
        return resp

    elif isinstance(resp, StreamHttpResponse):
        data = resp.stream.read()
        return HttpResponse(**{
            **{k: v for k, v in dataclass_shallow_asdict(resp).items() if k not in ('stream', '_closer')},
            'data': data,
        })

    else:
        raise TypeError(resp)


##


class HttpClient(Abstract):
    def __enter__(self: HttpClientT) -> HttpClientT:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def request(
            self,
            req: HttpRequest,
            *,
            check: bool = False,
    ) -> HttpResponse:
        with closing_response(self.stream_request(
                req,
                check=check,
        )) as resp:
            return read_response(resp)

    def stream_request(
            self,
            req: HttpRequest,
            *,
            check: bool = False,
    ) -> StreamHttpResponse:
        resp = self._stream_request(req)

        try:
            if check and not resp.is_success:
                if isinstance(resp.underlying, Exception):
                    cause = resp.underlying
                else:
                    cause = None
                raise HttpStatusError(read_response(resp)) from cause  # noqa

        except Exception:
            close_response(resp)
            raise

        return resp

    @abc.abstractmethod
    def _stream_request(self, req: HttpRequest) -> StreamHttpResponse:
        raise NotImplementedError



from .. import lang as _lang


with _lang.auto_proxy_init(globals()):
    ##

    from .bits import (  # noqa
        get_bit,
        get_bits,
        set_bit,
        set_bits,
    )

    from .c import (  # noqa
        cdiv,
        cmod,
    )

    from .fixed import (  # noqa
        CheckedFixedWidthIntError,
        OverflowFixedWidthIntError,
        UnderflowFixedWidthIntError,

        FixedWidthInt,

        SignedInt,
        UnsignedInt,

        CheckedInt,
        ClampedInt,

        AnyInt8,
        AnyInt16,
        AnyInt32,
        AnyInt64,
        AnyInt128,

        CheckedInt8,
        CheckedInt16,
        CheckedInt32,
        CheckedInt64,
        CheckedInt128,

        CheckedUint8,
        CheckedUint16,
        CheckedUint32,
        CheckedUint64,
        CheckedUint128,

        ClampedInt8,
        ClampedInt16,
        ClampedInt32,
        ClampedInt64,
        ClampedInt128,

        ClampedUint8,
        ClampedUint16,
        ClampedUint32,
        ClampedUint64,
        ClampedUint128,

        WrappedInt8,
        WrappedInt16,
        WrappedInt32,
        WrappedInt64,
        WrappedInt128,

        WrappedUint8,
        WrappedUint16,
        WrappedUint32,
        WrappedUint64,
        WrappedUint128,
    )

    from .floats import (  # noqa
        isclose,
        float_to_bytes,
        bytes_to_float,
    )

    from .stats import (  # noqa
        get_quantile,

        Stats,
    )

    from .histogram import (  # noqa
        SamplingHistogram,
    )
